# -*- coding: utf-8 -*-
"""
Test the tests.commons module
"""

from unittest import TestCase, expectedFailure

from . import commons


class FileCache(TestCase):
    """FileCache class test(s)"""

    @expectedFailure
    def test_checksum_error(self):
        """SHA256 checksum mismatch"""
        cache = commons.FileCache(commons.WRONG_CHECKSUM)
        # AssertionError due to checksum mismatch in the .raw property
        print(cache.raw)



# -*- coding: utf-8 -*-
"""
Test the supertabnanny.console module
"""

import itertools
import secrets

from unittest import TestCase

from supertabnanny import console


DARK_RANGE = range(0, 8)
BRIGHT_RANGE = range(60, 68)
VALID_COLOR_CODES = list(itertools.chain(DARK_RANGE, BRIGHT_RANGE))


class CsiMSequence(TestCase):
    """csi_m_sequence() function"""

    def test_number(self) -> None:
        """CSI-M-Sequence with a number"""
        self.assertEqual(console.csi_m_sequence(0), "\x1b[0m")

    def test_string(self) -> None:
        """CSI-M-Sequence with a string"""
        self.assertEqual(console.csi_m_sequence("31;47"), "\x1b[31;47m")


class CheckedColor(TestCase):
    """checked_color() function"""

    def test_valid(self) -> None:
        """valid values (0–7 or 60–67)"""
        for value in VALID_COLOR_CODES:
            with self.subTest("valid", value=value):
                self.assertEqual(console.checked_color(value), value)
            #
        #

    def test_invalid_fuzzed(self) -> None:
        """invalid values: everything else (fuzzed)"""
        not_inside_limits_re = r"not inside limits \(0, 7\) or \(60, 67\)"
        for iteration in range(100):
            value = 8 + secrets.randbelow(52)
            with self.subTest("invalid midrange", iteration=iteration, value=value):
                self.assertRaisesRegex(
                    ValueError,
                    f"^Color code {value} {not_inside_limits_re}$",
                    console.checked_color,
                    value,
                )
            #
        #
        for iteration in range(100):
            value = 68 + secrets.randbelow(188)
            with self.subTest("invalid high range", iteration=iteration, value=value):
                self.assertRaisesRegex(
                    ValueError,
                    f"^Color code {value} {not_inside_limits_re}$",
                    console.checked_color,
                    value,
                )
            #
        #
        for iteration in range(100):
            value = -1 - secrets.randbelow(2000)
            with self.subTest("invalid low range", iteration=iteration, value=value):
                self.assertRaisesRegex(
                    ValueError,
                    f"^Color code {value} {not_inside_limits_re}$",
                    console.checked_color,
                    value,
                )
            #
        #


class Bright(TestCase):
    """bright() function"""

    def test_valid(self) -> None:
        """valid values (0–7)"""
        for value in DARK_RANGE:
            with self.subTest("valid", value=value):
                self.assertEqual(console.bright(value), value + 60)
            #
        #

    def test_invalid_fuzzed(self) -> None:
        """invalid values: everything else (fuzzed)"""
        not_inside_limits_re = r"not inside limits \(0, 7\)"
        for iteration in range(100):
            value = 8 + secrets.randbelow(248)
            with self.subTest("invalid high range", iteration=iteration, value=value):
                self.assertRaisesRegex(
                    ValueError,
                    f"^Color code {value} {not_inside_limits_re}$",
                    console.bright,
                    value,
                )
            #
        #
        for iteration in range(100):
            value = -1 - secrets.randbelow(2000)
            with self.subTest("invalid low range", iteration=iteration, value=value):
                self.assertRaisesRegex(
                    ValueError,
                    f"^Color code {value} {not_inside_limits_re}$",
                    console.bright,
                    value,
                )
            #
        #


class BackgroudColor(TestCase):
    """background_color() function"""

    def test_none(self) -> None:
        """Test with None parameter"""
        self.assertIsNone(console.background_color(None))

    def test_valid(self) -> None:
        """Test with valid values"""
        for value in VALID_COLOR_CODES:
            with self.subTest("valid", value=value):
                self.assertEqual(console.background_color(value), value + 40)
            #
        #


class ForegroudColor(TestCase):
    """foreground_color() function"""

    def test_none(self) -> None:
        """Test with None parameter"""
        self.assertIsNone(console.foreground_color(None))

    def test_valid(self) -> None:
        """Test with valid values"""
        for value in VALID_COLOR_CODES:
            with self.subTest("valid", value=value):
                self.assertEqual(console.foreground_color(value), value + 30)
            #
        #


class EscapeSequenceStart(TestCase):
    """escape_sequence_start() function"""

    def test_no_color(self) -> None:
        """Test with no colors provided"""
        self.assertEqual(console.escape_sequence_start(), "")

    def test_bg_only(self) -> None:
        """Test with background color only"""
        for bg_color in VALID_COLOR_CODES:
            with self.subTest("valid", bg_color=bg_color):
                self.assertEqual(
                    console.escape_sequence_start(bg=bg_color),
                    f"\x1b[{bg_color + 40}m",
                )
            #
        #

    def test_fg_only(self) -> None:
        """Test with foreground color only"""
        for fg_color in VALID_COLOR_CODES:
            with self.subTest("valid", fg_color=fg_color):
                self.assertEqual(
                    console.escape_sequence_start(fg=fg_color),
                    f"\x1b[{fg_color + 30}m",
                )
            #
        #

    def test_all_combinations(self) -> None:
        """Test all valid foreground/background color combinations"""
        for fg_color, bg_color in itertools.product(
            VALID_COLOR_CODES, VALID_COLOR_CODES
        ):
            with self.subTest("valid", fg_color=fg_color, bg_color=bg_color):
                self.assertEqual(
                    console.escape_sequence_start(fg=fg_color, bg=bg_color),
                    f"\x1b[{fg_color + 30};{bg_color + 40}m",
                )
            #
        #


class Colorized(TestCase):
    """colorized() function"""

    def setUp(self) -> None:
        """Define an instance-wide end sequence"""
        self.end_fmt = "\x1b[0m"

    def test_no_color(self) -> None:
        """Test with no colors provided"""
        text = secrets.token_urlsafe()
        self.assertEqual(console.colorized(text), text)

    def test_bg_only(self) -> None:
        """Test with background color only"""
        text = secrets.token_urlsafe()
        for bg_color in VALID_COLOR_CODES:
            with self.subTest("valid", text=text, bg_color=bg_color):
                self.assertEqual(
                    console.colorized(text, bg=bg_color),
                    f"\x1b[{bg_color + 40}m{text}{self.end_fmt}",
                )
            #
        #

    def test_fg_only(self) -> None:
        """Test with foreground color only"""
        text = secrets.token_urlsafe()
        for fg_color in VALID_COLOR_CODES:
            with self.subTest("valid", text=text, fg_color=fg_color):
                self.assertEqual(
                    console.colorized(text, fg=fg_color),
                    f"\x1b[{fg_color + 30}m{text}{self.end_fmt}",
                )
            #
        #

    def test_all_combinations(self) -> None:
        """Test all valid foreground/background color combinations"""
        text = secrets.token_urlsafe()
        for fg_color, bg_color in itertools.product(
            VALID_COLOR_CODES, VALID_COLOR_CODES
        ):
            with self.subTest("valid", text=text, fg_color=fg_color, bg_color=bg_color):
                self.assertEqual(
                    console.colorized(text, fg=fg_color, bg=bg_color),
                    f"\x1b[{fg_color + 30};{bg_color + 40}m{text}{self.end_fmt}",
                )
            #
        #


class Arrow(TestCase):
    """arrow() function"""

    def test_deault_shape(self) -> None:
        """default shabe in various sizes"""
        tail = "\u2014"
        for size, expected_value in (
            (1, "⇥"),
            (2, f"{tail}⇥"),
            (4, f"{tail * 3}⇥"),
            (5, f"{tail * 4}⇥"),
        ):
            with self.subTest(
                "default shape", size=size, expected_value=expected_value
            ):
                self.assertEqual(console.arrow(size=size), expected_value)
            #
        #

    def test_custom_shape(self) -> None:
        """valid values (0–7)"""
        for size, head, tail, expected_value in (
            (3, ">", "=", "==>"),
            (7, "x", "+", "++++++x"),
        ):
            with self.subTest(
                "default shape",
                size=size,
                head=head,
                tail=tail,
                expected_value=expected_value,
            ):
                self.assertEqual(
                    console.arrow(size=size, head=head, tail=tail), expected_value
                )
            #
        #

    def test_invalid(self) -> None:
        """invalid size: 0 and below"""
        for size in range(-3, 1):
            with self.subTest("invalid", size=size):
                self.assertRaisesRegex(
                    ValueError,
                    "^Arrow size must be 1 or greater$",
                    console.arrow,
                    size=size,
                )
            #
        #



# -*- coding: utf-8 -*-

"""
supertabnanny.whitespace - whitespace analysis module
"""

import collections
import dataclasses
import unicodedata

from typing import ClassVar, Iterator, Type

from .console import (
    CYAN,
    EMPTY,
    SPACE_SYMBOL,
    END_FORMAT,
    arrow,
    escape_sequence_start,
)


_SP = "\x20"
_TAB = "\x09"
_LF = "\x0a"
_CR = "\x0d"

_COMMA_BLANK = ", "


@dataclasses.dataclass
class LineFragment:
    """Fragment of a line"""

    source: str

    def __str__(self) -> str:
        """String value"""
        return self.source

    def __eq__(self, other) -> bool:
        """String value"""
        return self.__class__ == other.__class__ and self.source == other.source


@dataclasses.dataclass
class WhitespaceFragment(LineFragment):
    """Whitespace fragment of a line: a single whitespace character"""

    source: str = EMPTY
    allow: ClassVar[str] = EMPTY

    def __post_init__(self) -> None:
        """check for valid source"""
        self.source = self.source or self.allow
        if len(self.source) > 1:
            raise ValueError(
                f"{self.__class__.__name__} instances must always contain"
                " a single character"
            )
        #
        if self.allow and self.source != self.allow:
            raise ValueError(
                f"{self.__class__.__name__} instances must always contain"
                f" {self.allow!r}, not {self.source!r}"
            )
        #
        if unicodedata.category(self.source) not in ("Cc", "Cf", "Zl", "Zp", "Zs"):
            raise ValueError(
                f"{self.source!r} does not appear to be a whitespace character"
            )
        #

    def __str__(self) -> str:
        """String value"""
        return repr(self.source)[1:-1]


@dataclasses.dataclass
class SpaceFragment(WhitespaceFragment):
    """Fragment of a line containing only spaces"""

    allow: ClassVar[str] = _SP

    def __str__(self) -> str:
        """String value"""
        return SPACE_SYMBOL


@dataclasses.dataclass
class TabFragment(WhitespaceFragment):
    """Fragment of a line containing a single Tab character"""

    allow: ClassVar[str] = _TAB

    def expanded_arrow(self, tabsize: int = 4, offset: int = 0) -> str:
        """expand the tab and replace it by an arrow of the expanded length"""
        expanded_size = tabsize - (offset % tabsize)
        return arrow(size=expanded_size)


@dataclasses.dataclass
class LineFeedFragment(WhitespaceFragment):
    """Fragment of a line containing a single line feed character"""

    allow: ClassVar[str] = _LF


@dataclasses.dataclass
class CarriageReturnFragment(WhitespaceFragment):
    """Fragment of a line containing a single carriage return character"""

    allow: ClassVar[str] = _CR


@dataclasses.dataclass
class Calculator:
    """Caculate indent characteristics"""

    source: str
    total: int = dataclasses.field(init=False, default=0)
    trailing_spaces: int = dataclasses.field(init=False, default=0)
    spaces_preceding_tabs: tuple[int, ...] = dataclasses.field(init=False, default=())
    frequencies: tuple[int, ...] = dataclasses.field(init=False, default=())

    def __post_init__(self) -> None:
        """Analyze source and store the analysis results"""
        if self.source:
            # current trailing spaces
            current_trailing_spaces: int = 0
            # collector for number spaces directly before tabs
            spaces_before_tabs: list[int] = []
            for position, character in enumerate(self.source):
                if character == _SP:
                    current_trailing_spaces += 1
                elif character == _TAB:
                    spaces_before_tabs.append(current_trailing_spaces)
                    current_trailing_spaces = 0
                else:
                    self.total = position
                    break
                #
            else:
                self.total = position + 1
            #
            self.trailing_spaces = current_trailing_spaces
            self.spaces_preceding_tabs = tuple(spaces_before_tabs)
        #
        if self.spaces_preceding_tabs:
            freq_counter = collections.Counter(self.spaces_preceding_tabs)
            freq: list[int] = []
            for runlength in range(max(freq_counter) + 1):
                freq.append(freq_counter[runlength])
            #
            self.frequencies = tuple(freq)
        #


@dataclasses.dataclass
class Witness:
    """Witness of indent sizes at a given tabsize"""

    tabsize: int
    own_level: int
    other_level: int


class LineFormatter:
    """Formatter for a line"""

    fragment_types: dict[str, Type[WhitespaceFragment]] = {
        _SP: SpaceFragment,
        _TAB: TabFragment,
        _LF: LineFeedFragment,
        _CR: CarriageReturnFragment,
    }

    def __init__(self, source: str) -> None:
        """Split the line into fragments"""
        self.fragments: tuple[LineFragment, ...] = tuple(self.__iter_fragments(source))
        # total number of tabs in the line
        self.ntabs = len(
            [
                fragment
                for fragment in self.fragments
                if isinstance(fragment, TabFragment)
            ]
        )

    def __iter_fragments(self, source: str) -> Iterator[LineFragment]:
        """Build the fragments"""
        collected: list[str] = []
        for character in source:
            if character == _SP or not character.isprintable():
                if collected:
                    yield LineFragment(EMPTY.join(collected))
                    collected.clear()
                #
                try:
                    yield self.fragment_types[character]()
                except KeyError:
                    yield WhitespaceFragment(character)
                #
            else:
                collected.append(character)
            #
        #
        if collected:
            yield LineFragment(EMPTY.join(collected))
        #

    @staticmethod
    def get_replacement(
        fragment: LineFragment, tabsize: int = 4, offset: int = 0
    ) -> str:
        """Return the replacement for a fragment"""
        if isinstance(fragment, TabFragment):
            return fragment.expanded_arrow(tabsize=tabsize, offset=offset)
        #
        return str(fragment)

    def iter_chunks(self, tabsize: int = 4) -> Iterator[tuple[bool, str]]:
        """Iterate over replacement chunks calculated for the given tabsize"""
        collected: list[str] = []
        last_whitespace = False
        current_whitespace = False
        offset = 0
        for fragment in self.fragments:
            current_whitespace = isinstance(fragment, WhitespaceFragment)
            if current_whitespace != last_whitespace:
                if collected:
                    yield last_whitespace, EMPTY.join(collected)
                    collected.clear()
                #
                last_whitespace = current_whitespace
            #
            replacement = self.get_replacement(fragment, tabsize=tabsize, offset=offset)
            collected.append(replacement)
            match fragment:
                case LineFeedFragment() | CarriageReturnFragment():
                    offset = 0
                case SpaceFragment():
                    offset += 1
                case _:
                    offset += len(replacement)
                #
            #
        #
        yield current_whitespace, EMPTY.join(collected)

    def format(
        self, tabsize: int = 4, fg: int | None = CYAN, bg: int | None = None
    ) -> str:
        """Return the line with whitespace characters replaced
        by their visualizations (spaces by middle dots, tabs by arrows,
        line feeds by \n, carriage return by \r, other whitespace
        characters by their backslash representation),
        and highlighted using ANSI escapes if fg and or bg are not None.
        """
        highlight_start = escape_sequence_start(fg=fg, bg=bg)
        highlight_end = END_FORMAT if highlight_start else EMPTY
        last_whitespace = False
        output_parts: list[str] = []
        for current_whitespace, chunk in self.iter_chunks(tabsize=tabsize):
            if current_whitespace and not last_whitespace:
                output_parts.append(f"{highlight_start}")
            elif last_whitespace and not current_whitespace:
                output_parts.append(f"{highlight_end}")
            #
            output_parts.append(chunk)
            last_whitespace = current_whitespace
        #
        if last_whitespace:
            output_parts.append(highlight_end)
        #
        return EMPTY.join(output_parts)


class WitnessesGroup:
    """Group of witnesses"""

    def __init__(self, prefix: str = "") -> None:
        """Initialization argument:

        *   _prefix_: prefix for the string representation
        """
        self.__prefix = prefix
        self.__witnesses: list[Witness] = []

    @property
    def tabsizes(self) -> list[int]:
        """Return the list of affected tabsizes"""
        return [witness.tabsize for witness in self.__witnesses]

    def add(self, witness: Witness) -> None:
        """Add _witness_"""
        self.__witnesses.append(witness)

    def __repr__(self) -> str:
        """Object representation"""
        return (
            f"{self.__class__.__name__}({self.__prefix!r})"
            f" with tabsizes {self.tabsizes}"
        )

    def __str__(self) -> str:
        """String representation"""
        ts = self.tabsizes
        if not ts:
            return self.__prefix
        #
        plural = "s" if len(ts) > 1 else ""
        return f"{self.__prefix}at tab size{plural} {_COMMA_BLANK.join(map(str, ts))}"


class Normalization:
    """Normalization of leading Whitespace in a string"""

    def __init__(self, source: str) -> None:
        """Determine leading whitespace in _source_,
        and store a normalized representation
        """
        self.__calculator = Calculator(source)
        self.__longest_run_of_spaces = max(
            len(self.frequencies) - 1, self.trailing_spaces
        )
        self.__explicit_tabs = sum(self.frequencies)
        self.__tabs_with_no_preceding_spaces: bool = len(self.frequencies) == 1

    # @staticmethod

    @property
    def explicit_tabs(self) -> int:
        """Number of explicit tab characters"""
        return self.__explicit_tabs

    @property
    def tabs_only(self) -> bool:
        """Special form 'T+': at least one tab, no spaces at all"""
        return self.__tabs_with_no_preceding_spaces and not self.trailing_spaces

    @property
    def tabs_with_no_preceding_spaces(self) -> bool:
        r"""Special form 'T+S\*': at least one tab, but no spaces before tabs"""
        return self.__tabs_with_no_preceding_spaces

    @property
    def longest_run_of_spaces(self) -> int:
        """Longest consecutive run of spaces"""
        return self.__longest_run_of_spaces

    @property
    def frequencies(self) -> tuple[int, ...]:
        r"""Frequencies of spaces directly preceding tabs,
        indexed by number of spaces before the tab character:

        *   frequencies\[0\]: number of tabs
            **not** directly preceded by spaces
        *   frequencies\[1\]: number of tabs
            directly preceded by **exactly 1** space
        *   frequencies\[2\]: number of tabs
            directly preceded by **excatly 2** spaces
        *   (...)
        *   frequencies\[_n_\]: number of tabs
            directly preceded by **excatly _n_** spaces
        """
        return self.__calculator.frequencies

    @property
    def trailing_spaces(self) -> int:
        """Number of trailing spaces"""
        return self.__calculator.trailing_spaces

    def __len__(self) -> int:
        """Total number of leading whitespace characters"""
        return self.__calculator.total

    def __eq__(self, other) -> bool:
        """Rich "equals": compare normalized repreesentation"""
        return (
            self.frequencies == other.frequencies
            and self.trailing_spaces == other.trailing_spaces
        )

    def get_expanded_size(self, tabsize: int) -> int:
        """Calculate expanded size depending on the provided _tabsize_

        1.  Calculate total number of tabs as the sum of explicit tabs
            and tabs inferred from leading spaces.

            *   base formula
                (ls = leading spaces, ts = tabsize, freq = frequencies):

                    tabs = explicit_tabs + sum(ls // ts * freq[ls])

                (the sum is calculated over all freq members)

            *   regarding that `ls // ts == 0` for all `ls < ts`,
                we can start the loop at `freq[ts]` instead of `freq[0]`

            *   An alternative approach would be to lift all indexes by ts
                (`increased_ls = ls + ts)`
                which would simplify the algorithm to

                    tabs = sum(increased_ls // ts * freq[ls])

                but involve more calculations, being less efficient
                than simply adding the known fixed number of explicit tabs

        2.  Calculate the expanded size by multiplying the total number of tabs
            with the tabsize and adding the trailing spaces
        """
        if tabsize < 1:
            raise ValueError("A tabsize of at lest 1 is required")
        #
        over_tabsize_frequencies = self.frequencies[tabsize:]
        additional_tabs = sum(
            leading_spaces // tabsize * frequency
            for leading_spaces, frequency in enumerate(
                over_tabsize_frequencies, start=tabsize
            )
        )
        total_tabs = self.__explicit_tabs + additional_tabs
        return total_tabs * tabsize + self.trailing_spaces


class LeadingWhitespace:
    """Handle leading whitespace

    An attempt to refactor the standard library’s tabnanny.Whitespace class
    """

    def __init__(self, source: str, compatibility_mode: bool = False) -> None:
        """Determine leading whitespace in source,
        assuming it contains only one line
        """
        self.raw = source
        self.__normalization = Normalization(source)
        self.__indent_by_tabsize: dict[int, int] = {}
        self.__compatibility_mode = compatibility_mode

    @property
    def no_spaces_before_tabs(self) -> bool:
        """Special form 'T*S*': no spaces before tabs,
        but maybe tabs and/or trailing spaces
        """
        return (
            self.normalization.tabs_with_no_preceding_spaces
            or not self.normalization.explicit_tabs
        )

    @property
    def longest_run_of_spaces(self) -> int:
        """Longest consecutive run of spaces"""
        return self.normalization.longest_run_of_spaces

    @property
    def normalization(self) -> Normalization:
        """normalization"""
        return self.__normalization

    @property
    def explicit_tabs(self) -> int:
        """number of tab characters"""
        return self.normalization.explicit_tabs

    def __len__(self) -> int:
        """Total number of characters"""
        return len(self.normalization)

    def __eq__(self, other) -> bool:
        """Rich "equals": compare normalized repreesentation"""
        return self.normalization == other.normalization

    def __ge__(self, other) -> bool:
        r"""Rich "greater or equal": at least for one tab size,
        the own expanded size is greater than or equal to the other one.

        This is always true for leading whitespace of the same length
        or greater when assuming a tab size of 1 (shorcut #1)

        In the frequent special case 'T\*S\*', the expanded size
        is always greater than or equal to the other instance
        if the number of tabs is greater (shortcut #2)
        """
        # shoercut #1
        if len(self) >= len(other):
            return True
        #
        if (
            self.normalization.tabs_with_no_preceding_spaces
            and other.normalization.tabs_with_no_preceding_spaces
            and self.explicit_tabs > other.explicit_tabs
        ):
            return True
        #
        for witness in self._iter_witnesses(other):
            if witness.own_level >= witness.other_level:
                return True
            #
        #
        return False

    def _iter_witnesses(self, other, start: int = 1) -> Iterator[Witness]:
        """Return an iterator over Witness instances"""
        max_tabsize = max(self.longest_run_of_spaces, other.longest_run_of_spaces) + 1
        for tabsize in range(start, max_tabsize + 1):
            yield Witness(
                tabsize,
                self.cached_expanded_size(tabsize),
                other.cached_expanded_size(tabsize),
            )
        #

    def greater_or_equal_witnesses(self, other) -> WitnessesGroup:
        """Return an iterator over Witness instances"""
        if self.__compatibility_mode:
            prefix = "indent not greater e.g. "
        else:
            prefix = "expanded size greater or equal "
        #
        group = WitnessesGroup(prefix)
        for witness in self._iter_witnesses(other):
            if witness.own_level >= witness.other_level:
                group.add(witness)
            #
        #
        return group

    def not_equal_witnesses(self, other) -> WitnessesGroup:
        """Return an iterator over Witness instances"""
        if self.__compatibility_mode:
            prefix = "indent not equal e.g. "
        else:
            prefix = "expanded size different "
        #
        group = WitnessesGroup(prefix)
        for witness in self._iter_witnesses(other):
            if witness.own_level != witness.other_level:
                group.add(witness)
            #
        #
        return group

    def cached_expanded_size(self, tabsize: int) -> int:
        """return cached expanded size by tabsize or calculate it"""
        try:
            return self.__indent_by_tabsize[tabsize]
        except KeyError:
            expanded = self.normalization.get_expanded_size(tabsize)
            self.__indent_by_tabsize[tabsize] = expanded
            return expanded
        #



# -*- coding: utf-8 -*-

"""
console colors and functions
"""

import logging


EMPTY = ""
LF = "\n"

ARROW_TAIL = chr(0x2014)
ARROW_HEAD = chr(0x21E5)
SPACE_SYMBOL = chr(0xB7)

# ANSI COLOR BASE NUMBERS
BLACK = 0
RED = 1
GREEN = 2
YELLOW = 3
BLUE = 4
MAGENTA = 5
CYAN = 6
WHITE = 7

_DARK_LIMITS = (BLACK, WHITE)
_BRIGHT_LIMITS = tuple(limit + 60 for limit in _DARK_LIMITS)

# COLOR_OFFSETS
_FOREGROUND_OFFSET = 30
_BACKGROUND_OFFSET = 40

CSI = "\x1b["

ROOT_LOGGER = logging.getLogger()


def initialize_root_logger(loglevel: int) -> None:
    """Initialize the root logger"""
    formatter = logging.Formatter("%(levelname)-8s | %(message)s")
    stream_handlers = [
        handler
        for handler in ROOT_LOGGER.handlers
        if isinstance(handler, logging.StreamHandler)
    ]
    if stream_handlers:
        current_handler = stream_handlers[0]
    else:
        current_handler = logging.StreamHandler()
        ROOT_LOGGER.addHandler(current_handler)
    #
    current_handler.setFormatter(formatter)
    current_handler.setLevel(loglevel)
    ROOT_LOGGER.setLevel(loglevel)


def csi_m_sequence(enclosed: str | int) -> str:
    r"""return a full CSI…m sequence"""
    return f"\x1b[{enclosed}m"


END_FORMAT = csi_m_sequence(0)


def checked_color(color: int) -> int:
    """Allow a dark or bright color"""
    if (
        _DARK_LIMITS[0] <= color <= _DARK_LIMITS[1]
        or _BRIGHT_LIMITS[0] <= color <= _BRIGHT_LIMITS[1]
    ):
        return color
    #
    raise ValueError(
        f"Color code {color} not inside limits {_DARK_LIMITS} or {_BRIGHT_LIMITS}"
    )


def bright(color: int) -> int:
    """bright color"""
    if _DARK_LIMITS[0] <= color <= _DARK_LIMITS[1]:
        return color + 60
    #
    raise ValueError(f"Color code {color} not inside limits {_DARK_LIMITS}")


def background_color(color: int | None) -> int | None:
    """Allow a dark or bright color"""
    if color is None:
        return color
    #
    return checked_color(color) + _BACKGROUND_OFFSET


def foreground_color(color: int | None) -> int | None:
    """Allow a dark or bright color"""
    if color is None:
        return color
    #
    return checked_color(color) + _FOREGROUND_OFFSET


def escape_sequence_start(fg: int | None = None, bg: int | None = None) -> str:
    """Return the escape sequence for setting fg and bg colors"""
    bg_color: int | None = background_color(bg)
    fg_color: int | None = foreground_color(fg)
    if bg_color is None:
        if fg_color is None:
            return EMPTY
        #
        return csi_m_sequence(int(fg_color))
    #
    if fg_color is None:
        return csi_m_sequence(int(bg_color))
    #
    return csi_m_sequence(f"{fg_color};{bg_color}")


def colorized(text: str, fg: int | None = None, bg: int | None = None) -> str:
    """Return text surrounded with ANSI excape sequences for color"""
    start_csi_sequence = escape_sequence_start(fg=fg, bg=bg)
    if not start_csi_sequence:
        return text
    #
    return f"{start_csi_sequence}{text}{END_FORMAT}"


def arrow(size: int = 1, head: str = ARROW_HEAD, tail: str = ARROW_TAIL) -> str:
    """Return a right-pointing console arrow of size _size_"""
    if size < 1:
        raise ValueError("Arrow size must be 1 or greater")
    #
    return EMPTY.join((tail * (size - 1), head))



# -*- coding: utf-8 -*-

"""
Command line script
"""

import argparse
import logging
import pathlib
import sys


from . import __version__
from .console import initialize_root_logger
from .frontend import OutputMode, check_dispatch


def parse_arguments(*test_args: str, test_context: bool = False) -> argparse.Namespace:
    """Parse commandline arguments"""
    main_parser = argparse.ArgumentParser(
        prog="supertabnanny",
        description="Detect ambiguous indentation in Python soure files",
    )
    main_parser.set_defaults(loglevel=logging.WARNING, output_mode=OutputMode.NORMAL)
    main_parser.add_argument(
        "--version",
        action="version",
        version=__version__,
        help="print version and exit",
    )
    lo_parser = main_parser.add_argument_group(
        "logging options",
        "control log output and partially also affect diagnostic output",
    )
    lo_mutex = lo_parser.add_mutually_exclusive_group()
    lo_mutex.add_argument(
        "-d",
        "--debug",
        action="store_const",
        const=logging.DEBUG,
        dest="loglevel",
        help="debug mode (message overkill)",
    )
    lo_mutex.add_argument(
        "-v",
        "--verbose",
        action="store_const",
        const=logging.INFO,
        dest="loglevel",
        help="verbose mode (print more messages than default)",
    )
    lo_mutex.add_argument(
        "-q",
        "--quiet",
        action="store_const",
        const=logging.ERROR,
        dest="loglevel",
        help="quiet mode: print error messages only."
        " In compatibility mode, print the file name only for each indentation error.",
    )
    dof_parser = main_parser.add_argument_group(
        "diagnostics format options", "change the diagnostic output format"
    )
    dof_mutex = dof_parser.add_mutually_exclusive_group()
    dof_mutex.add_argument(
        "-c",
        "--compatibility",
        action="store_const",
        const=OutputMode.COMPATIBILITY,
        dest="output_mode",
        help="same format as the original tabnanny module (compatibility mode)",
    )
    dof_mutex.add_argument(
        "-j",
        "--json",
        action="store_const",
        const=OutputMode.JSON,
        dest="output_mode",
        help="one JSON object per file (json mode)",
    )
    dof_mutex.add_argument(
        "-s",
        "--show-lines",
        type=int,
        nargs="?",
        metavar="NUMBER",
        const=3,
        help="show up to %(metavar)s sample lines before and including"
        " the affected line, with different tab sizes to illustrate the problem."
        " If specified without a number, %(const)s lines are shown.",
    )
    main_parser.add_argument(
        "--fail-fast",
        action="store_true",
        help="exit immediately if an exception is encountered",
    )
    main_parser.add_argument(
        "files_and_or_directories",
        type=pathlib.Path,
        metavar="FILE_OR_DIRECTORY",
        nargs="+",
        help="the file(s) or director{y|ies} to analyze",
    )
    args: tuple[str, ...] | None = None
    if test_context or test_args:
        args = test_args
    #
    arguments = main_parser.parse_args(args=args)
    initialize_root_logger(arguments.loglevel)
    return arguments


def main(*test_args: str, test_context: bool = False) -> int:
    """..."""
    arguments = parse_arguments(*test_args, test_context=test_context)
    success = True
    for target_path in arguments.files_and_or_directories:
        success &= check_dispatch(
            target_path,
            output_mode=arguments.output_mode,
            show_lines=arguments.show_lines,
            fail_fast=arguments.fail_fast,
        )
    #
    return 0 if success else 1


if __name__ == "__main__":
    sys.exit(main())



# -*- coding: utf-8 -*-

"""
frontend checks
"""

import logging
import pathlib
import sys
import textwrap

from enum import Enum
from typing import Iterator

from .console import LF, BLACK, WHITE, GREEN, RED, ROOT_LOGGER, colorized, bright
from .parser import IndentParser, Diagnosis, CleanDiagnosis, ExceptionDiagnosis
from .whitespace import LineFormatter


class OutputMode(Enum):
    """Output modes"""

    JSON = "JSON"
    COMPATIBILITY = "tabnanny-compatibility"
    NORMAL = "normal"


def frequent_tabsizes_iter() -> Iterator[int]:
    """Iterate over the most frequent tabsizes
    (own guess, not backed by any statistical data):
    4, 8, 2, 1, 6, 3, 5, 7, 9, 10, 11, …, 99
    """
    yield from (4, 8, 2, 1, 6, 3, 5, 7)
    yield from range(9, 100)


def iter_with_lineno(
    source_lines: list[str],
    lineno: int,
    max_lines: int = 3,
) -> Iterator[tuple[int, str]]:
    """Get a part of a source file"""
    start_line = 0
    if max_lines > 0:
        start_line = lineno - max_lines
        start_line = max(start_line, 0)
    #
    for index, line in enumerate(source_lines[start_line:lineno], start=start_line + 1):
        yield index, line
    #


# pylint: disable=too-many-locals ; refactoring candidate
def collect_samples(
    file_path: pathlib.Path,
    diagnosis: Diagnosis,
    max_lines: int = 3,
    samples_limit: int = 3,
) -> Iterator[tuple[int, str]]:
    """Collect samples from a file, each being max_lines long
    before and including line #lineno,
    and replace whitespace by symbols
    """
    if diagnosis.lineno < 1:
        return
    #
    source_lines = file_path.read_text(encoding="utf-8").splitlines(keepends=True)
    sample_formatters: list[tuple[int, LineFormatter]] = [
        (lineno, LineFormatter(line))
        for (lineno, line) in iter_with_lineno(
            source_lines, diagnosis.lineno, max_lines=max_lines
        )
    ]
    total_tabs = sum(formatter.ntabs for (_, formatter) in sample_formatters)
    if not total_tabs:
        samples_limit = 1
        logging.info(
            "No tabs in the sample, so it will not expand differently"
            " at varying tab sizes."
        )
    #
    frequent_tab_sizes = list(frequent_tabsizes_iter())
    remaining_tabsizes = set(diagnosis.tabsizes or frequent_tab_sizes)
    found_samples = 0
    for current_ts in frequent_tab_sizes:
        if found_samples >= samples_limit or not remaining_tabsizes:
            break
        #
        try:
            remaining_tabsizes.remove(current_ts)
        except KeyError:
            continue
        #
        found_samples += 1
        collected: list[str] = []
        for line_number, formatter in sample_formatters:
            lineno_prefix = colorized(f"{line_number:4d}: ", fg=bright(BLACK), bg=WHITE)
            collected.append(f"  {lineno_prefix}{formatter.format(tabsize=current_ts)}")
        #
        yield current_ts, LF.join(collected)
    #


def compatible_check_impl(
    file_path: str | pathlib.Path,
    fail_fast: bool = False,
) -> bool:
    """Check a file and produce the exacttly same output
    as the standard library tabnanny module
    """
    if isinstance(file_path, pathlib.Path):
        path_obj = file_path
    else:
        path_obj = pathlib.Path(file_path)
    #
    indent_parser = IndentParser(compatibility_mode=True)
    diagnosis = indent_parser.parse_file(path_obj)
    output = diagnosis.compatible_output()
    if output:
        if isinstance(diagnosis, ExceptionDiagnosis):
            sys.stderr.write(f"{output}\n")
        else:
            print(output)
        #
    #
    if fail_fast and isinstance(diagnosis, ExceptionDiagnosis):
        sys.exit(1)
    #
    return isinstance(diagnosis, CleanDiagnosis)


def compatible_walk_impl(
    file_or_directory: str | pathlib.Path,
) -> Iterator[pathlib.Path]:
    """Yield file names from the path, descending into directories only
    if they are not symbolic links
    """
    if isinstance(file_or_directory, pathlib.Path):
        path_obj = file_or_directory
    else:
        path_obj = pathlib.Path(file_or_directory)
    #
    if path_obj.is_file() and path_obj.suffix == ".py":
        yield path_obj
    elif path_obj.is_dir() and not path_obj.is_symlink():
        if ROOT_LOGGER.level < logging.WARNING:
            print(f"{str(path_obj)}: listing directory")
        #
        for sub_path in path_obj.glob("*"):
            if sub_path.is_dir() and not sub_path.is_symlink() or sub_path.is_file():
                yield from compatible_walk_impl(sub_path)
            #
        #
    #


def check_compatible(
    file_or_directory: str | pathlib.Path,
    fail_fast: bool = False,
) -> bool:
    """Check a file or directory and produce the exactly same output
    as the standard library tabnanny module – wrapper setting the
    verbose and filename_only flags and calling the walk and
    alling the tybnanny compatible walk and check implementations.
    """
    success = True
    for file_path in compatible_walk_impl(file_or_directory):
        success &= compatible_check_impl(
            file_path,
            fail_fast=fail_fast,
        )
    #
    return success


def display_samples(
    file_path: pathlib.Path,
    diagnosis: Diagnosis,
    max_lines: int = 0,
    line_width: int = 72,
) -> None:
    """Print samples of problematic indentation at varying tabsizes"""
    tabsizes_samples = list(collect_samples(file_path, diagnosis, max_lines=max_lines))
    if not tabsizes_samples:
        return
    #
    nsamples = len(tabsizes_samples)
    nlines = len(tabsizes_samples[0][1].splitlines())
    separator = "\u2500" * line_width
    print(f"\u256d{separator}\u256e")
    print(f"  {file_path}:")
    lines_description = (
        f"a {nlines} lines sample" if nlines > 1 else "the affected line"
    )
    if nsamples > 1:
        print(
            textwrap.fill(
                f"Showing {lines_description} expanded at {nsamples} different"
                " tabsizes to illustrate the problem",
                width=line_width,
                initial_indent="  ",
                subsequent_indent="  ",
            )
        )
    else:
        print(f"  Showing {lines_description} to illustrate the problem")
    for tabsize, sample in tabsizes_samples:
        print(f"  {f' expanded at tab size {tabsize} ':⸱^{line_width - 2}}")
        print(sample)
    #
    print(f"\u2570{separator}\u256f")
    # print(separator)


def check_newstyle(
    file_path: pathlib.Path,
    output_mode: OutputMode,
    show_lines: int | None = None,
    fail_fast: bool = False,
) -> bool:
    """check_newstyle(path, output_mode, show_lines)

    If file_or_dir is a directory and not a symbolic link, then recursively
    descend the directory tree named by file_or_dir, checking all .py files
    along the way. If file_or_dir is an ordinary Python source file, it is
    checked for whitespace related problems. The diagnostic messages are
    written to standard output using the print statement.
    """
    diagnostic_messages = ""
    diagnosis = IndentParser().parse_file(file_path)
    match output_mode:
        case OutputMode.JSON:
            print(diagnosis.json_output())
        case OutputMode.NORMAL:
            diagnostic_messages = diagnosis.normal_output()
            if diagnostic_messages:
                if diagnosis.success:
                    status_indicator = colorized(" \u2713 ", bg=GREEN)
                else:
                    status_indicator = colorized(" \u2717 ", bg=RED)
                #
                print(f"{status_indicator} {diagnostic_messages}")
            #
        #
    #
    if fail_fast and isinstance(diagnosis, ExceptionDiagnosis):
        logging.critical("Exiting on first error as requested")
        sys.exit(1)
    #
    if isinstance(show_lines, int):
        display_samples(file_path, diagnosis, max_lines=show_lines)
    #
    return diagnosis.success


def walk(file_or_directory: pathlib.Path) -> Iterator[pathlib.Path]:
    """Yield python file paths"""
    if file_or_directory.is_file() and file_or_directory.suffix.lower() == ".py":
        yield file_or_directory
    elif (
        file_or_directory.is_dir()
        and not file_or_directory.is_symlink()
        and not file_or_directory.name.startswith(".")
    ):
        logging.debug("descending into directory %s", file_or_directory)
        for subpath in file_or_directory.glob("*"):
            yield from walk(subpath)
        #
        logging.debug("back in directory %s", file_or_directory.parent)
    elif file_or_directory.exists():
        reason = " (symbolic link directory)" if file_or_directory.is_symlink() else ""
        logging.debug(
            "%s: ignored, not suitable for checking%s", file_or_directory, reason
        )
    else:
        logging.warning("%s: does not exist", file_or_directory)
    #


def check_dispatch(
    file_or_directory: pathlib.Path,
    output_mode: OutputMode,
    show_lines: int | None = None,
    fail_fast: bool = False,
) -> bool:
    """Dispatch checks to the compatible or newstyle functions"""
    if output_mode == OutputMode.COMPATIBILITY:
        return check_compatible(file_or_directory, fail_fast=fail_fast)
    #
    success = True
    for file_path in walk(file_or_directory):
        success &= check_newstyle(
            file_path,
            output_mode,
            show_lines=show_lines,
            fail_fast=fail_fast,
        )
    #
    return success



# -*- coding: utf-8 -*-
"""
Test parity with the standard library’s tabnanny module

"""

import contextlib
import io
import itertools
import logging
import sys
import tabnanny
import unittest

from typing import Any, Callable
from unittest.mock import patch

from supertabnanny import frontend

from . import commons


ALL_EXAMPLES = commons.AllSources()

NORMAL = "normal"
VERBOSE = "verbose"
QUIET = "quiet"
OUTPUT_LEVELS = NORMAL, VERBOSE, QUIET
STDOUT = "stdout"
STDERR = "stderr"
STREAM_NAMES: list[str] = [
    f"{prefix}_{suffix}"
    for prefix, suffix in itertools.product(OUTPUT_LEVELS, (STDOUT, STDERR))
]

RETURNCODE = "returncode"
RETURNVALUE = "returnvalue"
STN_ONLY_RETURNCODE = f"stn_only_{RETURNCODE}"

EMPTY = ""
NOTHING = object()

TABNANNY = "tabnanny"
SUPERTABNANNY = "supertabnanny"


@contextlib.contextmanager
def captured_output(stream_name):
    """Return a context manager used by captured_stdout/stdin/stderr
    that temporarily replaces the sys stream *stream_name* with a StringIO."""
    orig_stdout = getattr(sys, stream_name)
    setattr(sys, stream_name, io.StringIO())
    try:
        yield getattr(sys, stream_name)
    finally:
        setattr(sys, stream_name, orig_stdout)


def captured_stdout():
    """Capture the output of sys.stdout:

    with captured_stdout() as stdout:
        print("hello")
    self.assertEqual(stdout.getvalue(), "hello\\n")
    """
    return captured_output("stdout")


def captured_stderr():
    """Capture the output of sys.stderr:

    with captured_stderr() as stderr:
        print("hello", file=sys.stderr)
    self.assertEqual(stderr.getvalue(), "hello\\n")
    """
    return captured_output("stderr")


def wrap_func(func: Callable, *args, **kwargs) -> tuple[int | None, Any, str, str]:
    """Execute func(*args, **kwargs) and return
    returncode, returnvalue, stdout and stderr in a tuple
    """
    verbose = int(kwargs.pop("verbose", 0))
    filename_only = int(kwargs.pop("filename_only", 0))
    returncode: int | None = None
    returnvalue: Any = None
    with (
        patch("sys.exit") as mock_exit,
        captured_stdout() as stdout,
        captured_stderr() as stderr,
    ):
        with (
            patch("tabnanny.verbose", verbose),
            patch("tabnanny.filename_only", filename_only),
        ):
            returnvalue = func(*args, **kwargs)
    #
    if mock_exit.called:
        exit_args = mock_exit.call_args.args
        returncode = int(exit_args[0])
    #
    # pylint: disable=no-member ; streams temoporarily replaced by StringIO
    return returncode, returnvalue, stdout.getvalue(), stderr.getvalue()


def get_example(name: str, offending_line_number: int | None = None) -> tuple[str, str]:
    """Get an example, return a tuple containing the full filename as string
    and offending line content (offending_line_number as 1-based index)
    """
    cached_example = ALL_EXAMPLES[name]
    full_filename = str(cached_example.file_path)
    if offending_line_number is None:
        return full_filename, EMPTY
    #
    return full_filename, cached_example.get_line(offending_line_number)


class CheckFunctions(unittest.TestCase):
    """check() functions in tabnanny and supertabnanny"""

    def setUp(self) -> None:
        """Full diff"""
        # pylint: disable=invalid-name ; defined in unittest.TestCase
        self.maxDiff = None

    def _consistency_checked_results(
        self,
        results: dict[str, tuple[Any, Any]],
        backend: str = SUPERTABNANNY,
    ) -> dict[str, Any]:
        """check result data
        (ie. returncode and returnvalue tuples for each output mode)
        for consistency between all output modes
        """
        for first, second in itertools.combinations(OUTPUT_LEVELS, 2):
            with self.subTest(
                "return(code|value) consistency",
                backend=backend,
                first=first,
                second=second,
            ):
                self.assertEqual(results[first], results[second])
            #
        #
        normal_values = results[NORMAL]
        return {RETURNCODE: normal_values[0], RETURNVALUE: normal_values[1]}

    def _complete_checked_results(
        self,
        streams: dict[str, Any],
        results: dict[str, Any],
        backend: str = SUPERTABNANNY,
        **expected_results,
    ) -> dict[str, Any]:
        """check streams and returned data
        and return a dict containing all
        """
        checked_results = self._consistency_checked_results(results, backend=backend)
        for key in RETURNCODE, RETURNVALUE:
            expected: Any = expected_results.get(key, NOTHING)
            if NOTHING is expected:
                continue
            #
            with self.subTest("returned", backend=backend, key=key, expected=expected):
                self.assertEqual(checked_results[key], expected)
            #
        #
        for key in STREAM_NAMES:
            expected = expected_results.get(key, EMPTY)
            if NOTHING is expected:
                continue
            #
            with self.subTest(
                "stream contents", backend=backend, key=key, expected=expected
            ):
                self.assertEqual(streams[key], expected)
            #
        #
        return streams | checked_results

    def _tabnanny_check_file(
        self,
        file_name: str,
        **expected_results,
    ) -> dict[str, Any]:
        """..."""
        # tabnanny never provides a return value,
        # so put a possibly expected one aside and add it again after the check.
        # If a returncode for the supertabnanny only was specified with
        # stn_only_returncode, put that aside as well
        # and add it as expected returncode after the check
        expected_stn_results = {
            RETURNCODE: expected_results.pop(
                STN_ONLY_RETURNCODE, expected_results.get(RETURNCODE, NOTHING)
            ),
            RETURNVALUE: expected_results.pop(RETURNVALUE, NOTHING),
        }
        results: dict[str, tuple[Any, Any]] = {}
        streams: dict[str, Any] = {}
        for output_level in OUTPUT_LEVELS:
            verbose = 1 if output_level == VERBOSE else 0
            filename_only = 1 if output_level == QUIET else 0
            rc, rv, out, err = wrap_func(
                tabnanny.check, file_name, verbose=verbose, filename_only=filename_only
            )
            results.update({output_level: (rc, rv)})
            streams.update(
                {f"{output_level}_{STDOUT}": out, f"{output_level}_{STDERR}": err}
            )
        #
        return (
            self._complete_checked_results(
                streams, results, backend=TABNANNY, **expected_results
            )
            | expected_stn_results
        )

    def _stn_check_file_compatible(
        self,
        file_name: str,
        **expected_results,
    ) -> dict[str, Any]:
        """..."""
        results: dict[str, tuple[Any, Any]] = {}
        streams: dict[str, Any] = {}
        for output_level in OUTPUT_LEVELS:
            loglevel = logging.WARNING
            if output_level == VERBOSE:
                loglevel = logging.INFO
            elif output_level == QUIET:
                loglevel = logging.ERROR
            #
            with patch(
                "supertabnanny.parser.ROOT_LOGGER",
                new=commons.MockLogger(level=loglevel),
            ):
                rc, rv, out, err = wrap_func(
                    frontend.check_compatible,
                    file_name,
                    fail_fast=True,
                )
            #
            results.update({output_level: (rc, rv)})
            streams.update(
                {f"{output_level}_{STDOUT}": out, f"{output_level}_{STDERR}": err}
            )
        #
        return self._complete_checked_results(
            streams, results, backend=SUPERTABNANNY, **expected_results
        )

    def test_error_free_file(self) -> None:
        """test with the error free file
        that triggers an IndentationError exception
        """
        file_name = get_example(commons.ERROR_FREE)[0]
        tn_results = self._tabnanny_check_file(
            file_name,
            # Expected results:
            verbose_stdout=f"{file_name!r}: Clean bill of health.\n",
            returnvalue=True,
            returncode=None,
        )
        # Compare tn_results to supertabnanny results
        self._stn_check_file_compatible(file_name, **tn_results)

    @unittest.skipIf(sys.version_info[:2] >= (3, 12), "up to 3.11 only")
    def test_incomplete_expression_file_up_to_311(self) -> None:
        """test with the incomplete-expression file
        that triggers a TokenError exception
        """
        file_name = get_example(commons.INCOMPLETE_EXPRESSION)[0]
        expected_message = (
            f"{file_name!r}: Token Error: ('EOF in multi-line statement', (9, 0))\n"
        )
        tn_results = self._tabnanny_check_file(
            file_name,
            # Expected results:
            normal_stderr=expected_message,
            verbose_stderr=expected_message,
            quiet_stderr=expected_message,
            returnvalue=False,
            stn_only_returncode=1,
        )
        # Compare tn_results to supertabnanny results
        self._stn_check_file_compatible(file_name, **tn_results)

    @unittest.skipIf(sys.version_info[:2] <= (3, 11), "from 3.12 only")
    def test_incomplete_expression_file_from_312(self) -> None:
        """test with the incomplete-expression file
        that triggers a TokenError exception
        """
        file_name = get_example(commons.INCOMPLETE_EXPRESSION)[0]
        expected_message = (
            f"{file_name!r}: Token Error:"
            " ('unexpected EOF in multi-line statement', (8, 0))\n"
        )
        tn_results = self._tabnanny_check_file(
            file_name,
            # Expected results:
            normal_stderr=expected_message,
            verbose_stderr=expected_message,
            quiet_stderr=expected_message,
            returnvalue=False,
            returncode=1,
        )
        # Compare tn_results to supertabnanny results
        self._stn_check_file_compatible(file_name, **tn_results)

    @unittest.skipIf(sys.version_info[:2] >= (3, 12), "up to 3.11 only")
    def test_wrong_indented_file_up_to_311(self) -> None:
        """test with the wrong-indented file
        that triggers an IndentationError exception
        """
        offending_line_number = 3
        file_name = get_example(commons.WRONG_INDENTED)[0]
        err_loc = "tokenize"
        expected_message = (
            f"{file_name!r}: Indentation Error: unindent does not match any"
            f" outer indentation level (<{err_loc}>, line {offending_line_number})\n"
        )
        tn_results = self._tabnanny_check_file(
            file_name,
            # Expected results:
            normal_stderr=expected_message,
            verbose_stderr=expected_message,
            quiet_stderr=expected_message,
            returnvalue=False,
            stn_only_returncode=1,
        )
        # Compare tn_results to supertabnanny results
        self._stn_check_file_compatible(file_name, **tn_results)

    @unittest.skipIf(sys.version_info[:2] <= (3, 11), "from 3.12 only")
    def test_wrong_indented_file_from_312(self) -> None:
        """test with the wrong-indented file
        that triggers an IndentationError exception
        """
        offending_line_number = 3
        file_name = get_example(commons.WRONG_INDENTED)[0]
        err_loc = "string"
        expected_message = (
            f"{file_name!r}: Indentation Error: unindent does not match any"
            f" outer indentation level (<{err_loc}>, line {offending_line_number})\n"
        )
        tn_results = self._tabnanny_check_file(
            file_name,
            # Expected results:
            normal_stderr=expected_message,
            verbose_stderr=expected_message,
            quiet_stderr=expected_message,
            returnvalue=False,
            returncode=1,
        )
        # Compare tn_results to supertabnanny results
        self._stn_check_file_compatible(file_name, **tn_results)

    @unittest.skipIf(sys.version_info[:2] >= (3, 12), "up to 3.11 only")
    def test_inequal_indent_file_up_to_311(self) -> None:
        """test with the not-equal file that triggers a NannyNag exception"""
        offending_line_number = 3
        file_name, offending_line = get_example(
            commons.NOT_EQUAL, offending_line_number=offending_line_number
        )
        verbose_reason = "indent not equal e.g. at tab sizes 1, 2, 3, 4, 5"
        tn_results = self._tabnanny_check_file(
            file_name,
            # Expected results:
            normal_stdout=f"{file_name} 3 {offending_line!r}\n",
            verbose_stdout=(
                f"{file_name!r}: *** Line 3: trouble in tab city! ***\n"
                f"offending line: {offending_line!r}\n{verbose_reason}\n"
            ),
            quiet_stdout=f"{file_name}\n",
            returncode=None,
            stn_only_returnvalue=False,
        )
        # Compare tn_results to supertabnanny results
        self._stn_check_file_compatible(file_name, **tn_results)

    @unittest.skipIf(sys.version_info[:2] <= (3, 11), "from 3.12 only")
    def test_inequal_indent_file_from_312(self) -> None:
        """test with the not-equal file that triggers a NannyNag exception"""
        offending_line_number = 3
        file_name, offending_line = get_example(
            commons.NOT_EQUAL, offending_line_number=offending_line_number
        )
        verbose_reason = commons.TAB_ERROR_MESSAGE
        tn_results = self._tabnanny_check_file(
            file_name,
            # Expected results:
            normal_stdout=f"{file_name} 3 {offending_line!r}\n",
            verbose_stdout=(
                f"{file_name!r}: *** Line 3: trouble in tab city! ***\n"
                f"offending line: {offending_line!r}\n{verbose_reason}\n"
            ),
            quiet_stdout=f"{file_name}\n",
            returncode=None,
            stn_only_returnvalue=False,
        )
        # Compare tn_results to supertabnanny results
        self._stn_check_file_compatible(file_name, **tn_results)

    @unittest.skipIf(sys.version_info[:2] >= (3, 12), "up to 3.11 only")
    def test_nannynag_errored_file_up_to_311(self) -> None:
        """test with the nannaynag-errorred file that triggers a NannyNag exception"""
        offending_line_number = 3
        file_name, offending_line = get_example(
            commons.NANNYNAG_ERRORED, offending_line_number=offending_line_number
        )
        verbose_reason = "indent not equal e.g. at tab size 1"
        tn_results = self._tabnanny_check_file(
            file_name,
            # Expected results:
            normal_stdout=f"{file_name} 3 {offending_line!r}\n",
            verbose_stdout=(
                f"{file_name!r}: *** Line 3: trouble in tab city! ***\n"
                f"offending line: {offending_line!r}\n{verbose_reason}\n"
            ),
            quiet_stdout=f"{file_name}\n",
            returncode=None,
            stn_only_returnvalue=False,
        )
        # Compare tn_results to supertabnanny results
        self._stn_check_file_compatible(file_name, **tn_results)

    @unittest.skipIf(sys.version_info[:2] <= (3, 11), "from 3.12 only")
    def test_nannynag_errored_file_from_312(self) -> None:
        """test with the nannaynag-errorred file that triggers a NannyNag exception"""
        offending_line_number = 3
        file_name, offending_line = get_example(
            commons.NANNYNAG_ERRORED, offending_line_number=offending_line_number
        )
        verbose_reason = commons.TAB_ERROR_MESSAGE
        tn_results = self._tabnanny_check_file(
            file_name,
            # Expected results:
            normal_stdout=f"{file_name} 3 {offending_line!r}\n",
            verbose_stdout=(
                f"{file_name!r}: *** Line 3: trouble in tab city! ***\n"
                f"offending line: {offending_line!r}\n{verbose_reason}\n"
            ),
            quiet_stdout=f"{file_name}\n",
            returncode=None,
            stn_only_returnvalue=False,
        )
        # Compare tn_results to supertabnanny results
        self._stn_check_file_compatible(file_name, **tn_results)

    @unittest.skipIf(sys.version_info[:2] >= (3, 12), "up to 3.11 only")
    def test_tab_space_error_file_1_up_to_311(self) -> None:
        """test with the ...  that triggers a NannyNag exception"""
        offending_line_number = 4
        file_name, offending_line = get_example(
            commons.TAB_SPACE_ERRORED_1, offending_line_number=offending_line_number
        )
        verbose_reason = "indent not greater e.g. at tab sizes 1, 2"
        tn_results = self._tabnanny_check_file(
            file_name,
            # Expected results:
            normal_stdout=f"{file_name} {offending_line_number} {offending_line!r}\n",
            verbose_stdout=(
                f"{file_name!r}: *** Line {offending_line_number}:"
                " trouble in tab city! ***\n"
                f"offending line: {offending_line!r}\n{verbose_reason}\n"
            ),
            quiet_stdout=f"{file_name}\n",
            returncode=None,
            stn_only_returnvalue=False,
        )
        # Compare tn_results to supertabnanny results
        self._stn_check_file_compatible(file_name, **tn_results)

    @unittest.skipIf(sys.version_info[:2] <= (3, 11), "from 3.12 only")
    def test_tab_space_error_file_1_from_312(self) -> None:
        """test with the ... that triggers a NannyNag exception"""
        offending_line_number = 4
        file_name, offending_line = get_example(
            commons.TAB_SPACE_ERRORED_1, offending_line_number=offending_line_number
        )
        verbose_reason = commons.TAB_ERROR_MESSAGE
        tn_results = self._tabnanny_check_file(
            file_name,
            # Expected results:
            normal_stdout=f"{file_name} {offending_line_number} {offending_line!r}\n",
            verbose_stdout=(
                f"{file_name!r}: *** Line {offending_line_number}:"
                " trouble in tab city! ***\n"
                f"offending line: {offending_line!r}\n{verbose_reason}\n"
            ),
            quiet_stdout=f"{file_name}\n",
            returncode=None,
            stn_only_returnvalue=False,
        )
        # Compare tn_results to supertabnanny results
        self._stn_check_file_compatible(file_name, **tn_results)

    @unittest.skipIf(sys.version_info[:2] >= (3, 12), "up to 3.11 only")
    def test_tab_space_error_file_2_up_to_311(self) -> None:
        """test with the ...  that triggers a NannyNag exception"""
        offending_line_number = 4
        file_name, offending_line = get_example(
            commons.TAB_SPACE_ERRORED_2, offending_line_number=offending_line_number
        )
        verbose_reason = "indent not equal e.g. at tab sizes 1, 2, 3, 4, 5, 6, 7, 9"
        tn_results = self._tabnanny_check_file(
            file_name,
            # Expected results:
            normal_stdout=f"{file_name} {offending_line_number} {offending_line!r}\n",
            verbose_stdout=(
                f"{file_name!r}: *** Line {offending_line_number}:"
                " trouble in tab city! ***\n"
                f"offending line: {offending_line!r}\n{verbose_reason}\n"
            ),
            quiet_stdout=f"{file_name}\n",
            returncode=None,
            stn_only_returnvalue=False,
        )
        # Compare tn_results to supertabnanny results
        self._stn_check_file_compatible(file_name, **tn_results)

    @unittest.skipIf(sys.version_info[:2] <= (3, 11), "from 3.12 only")
    def test_tab_space_error_file_2_from_312(self) -> None:
        """test with the ... that triggers a NannyNag exception"""
        offending_line_number = 4
        file_name, offending_line = get_example(
            commons.TAB_SPACE_ERRORED_2, offending_line_number=offending_line_number
        )
        verbose_reason = commons.TAB_ERROR_MESSAGE
        tn_results = self._tabnanny_check_file(
            file_name,
            # Expected results:
            normal_stdout=f"{file_name} {offending_line_number} {offending_line!r}\n",
            verbose_stdout=(
                f"{file_name!r}: *** Line {offending_line_number}:"
                " trouble in tab city! ***\n"
                f"offending line: {offending_line!r}\n{verbose_reason}\n"
            ),
            quiet_stdout=f"{file_name}\n",
            returncode=None,
            stn_only_returnvalue=False,
        )
        # Compare tn_results to supertabnanny results
        self._stn_check_file_compatible(file_name, **tn_results)



# -*- coding: utf-8 -*-
"""
Test the supertabnanny.whitespace module
"""

# import itertools
import secrets

from unittest import TestCase

from supertabnanny import whitespace

EMPTY = ""
SP = " "
TAB = "\t"
SP_TAB = f"{SP}{TAB}"
LETTER_S = "S"
LETTER_T = "T"
LETTERS_ST = f"{LETTER_S}{LETTER_T}"
SYMBOLIZE = str.maketrans(SP_TAB, LETTERS_ST)
DESYMBOLIZE = str.maketrans(LETTERS_ST, SP_TAB)

KW_NUMBER_OF_TABS = "number of tabs"
KW_TOTAL_NUMBER_OF_CHARS = "total number"
KW_NO_TABS = "not tabs"
KW_TABS_ONLY = "tabs only"
KW_LONGEST_RUN_OF_SPACES = "longest run of spaces"
KW_INDENT_LEVEL = "index level"
KW_NO_SBT = "no spaces before tabs"
KW_EXPANDED_SIZE = "expanded size"
KW_ATTRIBUTE = "attribute"
KW_FREQUENCIES = "frequencies"
KW_TRAILING_SPACES = "trailing spaces"

WG_GE_COMPATIBILITY = "indent not greater e.g. "
WG_GE_NEWSTYLE = "expanded size greater or equal "
WG_NE_COMPATIBILITY = "indent not equal e.g. "
WG_NE_NEWSTYLE = "expanded size different "


FUZZ_ITERATIONS = 500

#
# Helper functions
#


def get_random_prefix(min_length: int = 2, max_length: int = 80) -> str:
    """Return a random combination of tabs and spaces between the two lengths"""
    length = min_length + secrets.randbelow(max_length + 1 - min_length)
    collector = [secrets.choice(SP_TAB) for _ in range(length)]
    return EMPTY.join(collector)


def sym2ws(symbolic: str) -> str:
    """Convert symbols S and T for spaces and tabs in a string to whitespace"""
    return symbolic.translate(DESYMBOLIZE)


def ws2sym(preceding_whitespace: str) -> str:
    """Convert whitespace in a string to symbols S and T for spaces and Tabs"""
    return preceding_whitespace.translate(SYMBOLIZE)


#
# Test cases
#


class LineFragment(TestCase):
    """LineFragment class"""

    def test_str(self) -> None:
        """Initialization and string representation"""
        for source in ("Ole", "Dole", "Doffen"):
            with self.subTest("str", source=source):
                self.assertEqual(str(whitespace.LineFragment(source)), source)
            #
        #

    def test_eq(self) -> None:
        """equality"""
        with self.subTest("equal"):
            self.assertEqual(
                whitespace.LineFragment("Ole"),
                whitespace.LineFragment("Ole"),
            )
        #
        with self.subTest("equal"):
            self.assertNotEqual(
                whitespace.LineFragment("Dole"),
                whitespace.LineFragment("Doffen"),
            )
        #


class WhitespaceFragment(TestCase):
    """WhitespaceFragment class"""

    def test_postinit_errors(self) -> None:
        """__post_init__() method"""
        for source, expected_error_message_re in (
            (
                "  ",
                "^WhitespaceFragment instances must always contain a single character$",
            ),
            (
                "x",
                "^'x' does not appear to be a whitespace character",
            ),
        ):
            with self.subTest("__post_init__", source=source):
                self.assertRaisesRegex(
                    ValueError,
                    expected_error_message_re,
                    whitespace.WhitespaceFragment,
                    source,
                )
            #
        #

    def test_str(self) -> None:
        """Initialization and string representation"""
        testcases: list[tuple[str, str]] = [
            ("\t", "\\t"),
            ("\n", "\\n"),
            ("\r", "\\r"),
            ("\x0b", "\\x0b"),
            ("\x0c", "\\x0c"),
        ]
        for codepoint in (
            0x85,
            0xA0,
            0x1680,
            0x180E,
            *(range(0x2000, 0x200E)),
            0x2028,
            0x2029,
            0x202F,
            0x205F,
            0x2060,
            0x3000,
            0xFEFF,
        ):
            character = chr(codepoint)
            testcases.append(
                (
                    character,
                    character.encode("ascii", errors="backslashreplace").decode(
                        "utf-8"
                    ),
                )
            )
        for source, expected_value in testcases:
            with self.subTest("str", source=source):
                self.assertEqual(
                    str(whitespace.WhitespaceFragment(source)), expected_value
                )
            #
        #


class SpaceFragment(TestCase):
    """SpaceFragment class"""

    def test_postinit_errors(self) -> None:
        """__post_init__() method"""
        for source, expected_error_message_re in (
            (
                "\t",
                "^SpaceFragment instances must always contain ' ', not '\\\\t'$",
            ),
            (
                "y",
                "^SpaceFragment instances must always contain ' ', not 'y'$",
            ),
        ):
            with self.subTest("__post_init__", source=source):
                self.assertRaisesRegex(
                    ValueError,
                    expected_error_message_re,
                    whitespace.SpaceFragment,
                    source,
                )
            #
        #

    def test_str(self) -> None:
        """Initialization and string representation"""
        self.assertEqual(str(whitespace.SpaceFragment()), "·")


class TabFragment(TestCase):
    """TabFragment class"""

    def test_postinit_errors(self) -> None:
        """__post_init__() method"""
        self.assertRaisesRegex(
            ValueError,
            "^TabFragment instances must always contain '\\\\t', not '\\\\n'$",
            whitespace.TabFragment,
            "\n",
        )

    def test_str(self) -> None:
        """Initialization and string representation"""
        self.assertEqual(str(whitespace.TabFragment()), "\\t")

    def test_expanded_arrow(self) -> None:
        """expanded_arrow() method"""
        tab_fragment = whitespace.TabFragment()
        for arrow_data, expected_result in (
            ({"tabsize": 4, "offset": 0}, "———⇥"),
            ({"tabsize": 5, "offset": 7}, "——⇥"),
            ({"tabsize": 1, "offset": 0}, "⇥"),
            ({"tabsize": 8, "offset": 1}, "——————⇥"),
        ):
            with self.subTest("expanded_arror", expected_result=expected_result):
                self.assertEqual(
                    tab_fragment.expanded_arrow(**arrow_data), expected_result
                )
            #
        #


class LineFeedFragment(TestCase):
    """LineFeedFragment class"""

    def test_postinit_errors(self) -> None:
        """__post_init__() method"""
        self.assertRaisesRegex(
            ValueError,
            "^LineFeedFragment instances must always contain '\\\\n', not '\\\\r'$",
            whitespace.LineFeedFragment,
            "\r",
        )

    def test_str(self) -> None:
        """Initialization and string representation"""
        self.assertEqual(str(whitespace.LineFeedFragment()), "\\n")


class CarriageReturnFragment(TestCase):
    """CarriageReturnFragment class"""

    def test_postinit_errors(self) -> None:
        """__post_init__() method"""
        self.assertRaisesRegex(
            ValueError,
            "^CarriageReturnFragment instances must always contain"
            " '\\\\r', not '\\\\t'$",
            whitespace.CarriageReturnFragment,
            "\t",
        )

    def test_str(self) -> None:
        """Initialization and string representation"""
        self.assertEqual(str(whitespace.CarriageReturnFragment()), "\\r")


class Witness(TestCase):
    """Witness class"""

    def test_attributes(self) -> None:
        """Initialization and attributes"""
        tabsize, own_level, other_level = 4, 7, 8
        witness = whitespace.Witness(tabsize, own_level, other_level)
        with self.subTest(KW_ATTRIBUTE, name="tabsize"):
            self.assertEqual(witness.tabsize, tabsize)
        #
        with self.subTest(KW_ATTRIBUTE, name="own_level"):
            self.assertEqual(witness.own_level, own_level)
        #
        with self.subTest(KW_ATTRIBUTE, name="other_level"):
            self.assertEqual(witness.other_level, other_level)
        #


class WitnessesGroup(TestCase):
    """WitnessesGroup class"""

    def test_single(self) -> None:
        """group with only a single witness"""
        witnesses_group = whitespace.WitnessesGroup(prefix="single witness ")
        witnesses_group.add(whitespace.Witness(1, 2, 3))
        self.assertEqual(str(witnesses_group), "single witness at tab size 1")

    def test_multiple(self) -> None:
        """group with muttiple witnesses"""
        witnesses_group = whitespace.WitnessesGroup(prefix="multiple witnesses ")
        witnesses_group.add(whitespace.Witness(4, 5, 6))
        witnesses_group.add(whitespace.Witness(7, 8, 9))
        witnesses_group.add(whitespace.Witness(20, 5, 5))
        self.assertEqual(
            str(witnesses_group), "multiple witnesses at tab sizes 4, 7, 20"
        )


class LineFormatter(TestCase):
    """LineFormatter class"""

    def setUp(self) -> None:
        """Full diff"""
        # pylint: disable=invalid-name ; defined in unittest.TestCase
        self.maxDiff = None

    def test_attributes(self) -> None:
        """Initialization and attributes"""
        for source, expected_ntabs, expected_fragments in (
            (
                " \t print('hello world')\r\n",
                1,
                (
                    whitespace.SpaceFragment(" "),
                    whitespace.TabFragment("\t"),
                    whitespace.SpaceFragment(" "),
                    whitespace.LineFragment("print('hello"),
                    whitespace.SpaceFragment(" "),
                    whitespace.LineFragment("world')"),
                    whitespace.LineFeedFragment("\n"),
                ),
            ),
        ):
            formatter = whitespace.LineFormatter(source)
            with self.subTest(
                "attributes", source=source, expected_ntabs=expected_ntabs
            ):
                self.assertEqual(formatter.ntabs, expected_ntabs)
            #
            with self.subTest(
                "attributes", source=source, expected_fragments=expected_fragments
            ):
                self.assertEqual(formatter.ntabs, expected_ntabs)
            #
        #

    def test_arrow_head_positions(self):
        """.format() method: test if all arrow heads of tab replacements
        are at correct positions (each directly before a tab stop,
        position % tabsize == tabsize - 1)
        """
        for source in (
            " \t print('hello world')\r\n",
            "\t\t\tprint('hello\tworld')\n",
            " \t  \t   \tprint('\thello \t       \t   \tworld\t\u2000')",
        ):
            for tabsize in range(1, 100):
                formatted = whitespace.LineFormatter(source).format(
                    tabsize=tabsize, fg=None, bg=None
                )
                found_at = -1
                while True:
                    found_at = formatted.find(chr(0x21E5), found_at + 1)
                    if found_at == -1:
                        break
                    #
                    with self.subTest(
                        "arrow head", source=source, tabsize=tabsize, found_at=found_at
                    ):
                        self.assertEqual(found_at % tabsize, tabsize - 1)
                    #
                #
            #
        #


class Normalization(TestCase):
    """LeadingWhitespace class"""

    def test_attributes_fuzzed(self) -> None:
        """initialization with various combinations of spaces and tabs"""
        for iteration in range(FUZZ_ITERATIONS):
            source = get_random_prefix()
            normal = whitespace.Normalization(source)
            display = ws2sym(source)
            with self.subTest(
                display,
                iteration=iteration,
                subject=KW_NUMBER_OF_TABS,
            ):
                self.assertEqual(normal.explicit_tabs, source.count(TAB))
            #
            with self.subTest(
                display,
                iteration=iteration,
                subject=KW_TOTAL_NUMBER_OF_CHARS,
            ):
                self.assertEqual(len(normal), len(source))
            #
            with self.subTest(
                display,
                iteration=iteration,
                subject=KW_TABS_ONLY,
            ):
                self.assertEqual(normal.tabs_only, SP not in source)
            #
            with self.subTest(
                display,
                iteration=iteration,
                subject=KW_LONGEST_RUN_OF_SPACES,
            ):
                self.assertEqual(
                    normal.longest_run_of_spaces,
                    max(len(item) for item in source.split(TAB)),
                )
            #
        #

    def test_attributes(self) -> None:
        """initialization and attributes for a selected source"""
        symbolic_prefix = "SSSSSTSSSSSSSSSTSSTSTTTSSTTSSSSSSSSTSSS"
        source = f"{sym2ws(symbolic_prefix)}real_line_content = True"
        normal = whitespace.Normalization(source)
        with self.subTest(
            symbolic_prefix,
            subject=KW_FREQUENCIES,
        ):
            self.assertEqual(normal.frequencies, (3, 1, 2, 0, 0, 1, 0, 0, 1, 1))
        #
        with self.subTest(
            symbolic_prefix,
            subject=KW_TRAILING_SPACES,
        ):
            self.assertEqual(normal.trailing_spaces, 3)
        #
        with self.subTest(
            symbolic_prefix,
            subject=KW_NUMBER_OF_TABS,
        ):
            self.assertEqual(normal.explicit_tabs, symbolic_prefix.count(LETTER_T))
        #
        with self.subTest(
            symbolic_prefix,
            subject=KW_TOTAL_NUMBER_OF_CHARS,
        ):
            self.assertLess(len(normal), len(source))
            self.assertEqual(len(normal), len(symbolic_prefix))
        #
        with self.subTest(
            symbolic_prefix,
            subject=KW_TABS_ONLY,
        ):
            self.assertFalse(normal.tabs_only)
        #
        with self.subTest(
            symbolic_prefix,
            subject=KW_LONGEST_RUN_OF_SPACES,
        ):
            self.assertEqual(
                normal.longest_run_of_spaces,
                max(len(item) for item in symbolic_prefix.split(LETTER_T)),
            )
        #

    def test_frequencies(self) -> None:
        """.frequencies property"""
        for symbolic_source, expected_frequencies in (
            ("SSTTSTSSST", (1, 1, 1, 1)),
            ("STSTSTTTSS", (2, 3)),
        ):
            with self.subTest("norm", symbolic_source=symbolic_source):
                self.assertEqual(
                    whitespace.Normalization(sym2ws(symbolic_source)).frequencies,
                    expected_frequencies,
                )
            #
        #

    def test_expanded_size_error(self) -> None:
        """.get_expanded_size() method, errors with invalid tab sizes"""
        normalization = whitespace.Normalization("example")
        for invalid_tab_size in (-2, -1, 0):
            with self.subTest("invalid tab_size", invalid_tab_size=invalid_tab_size):
                self.assertRaisesRegex(
                    ValueError,
                    "^A tabsize of at lest 1 is required$",
                    normalization.get_expanded_size,
                    invalid_tab_size,
                )
            #
        #

    def test_expanded_size_fuzzed(self) -> None:
        """.get_expanded_size() method, fuzzed"""
        for iteration in range(FUZZ_ITERATIONS):
            source = get_random_prefix()
            normal = whitespace.Normalization(source)
            display = ws2sym(source)
            for tab_size in range(1, normal.longest_run_of_spaces + 1):
                with self.subTest(
                    display,
                    iteration=iteration,
                    subject=KW_EXPANDED_SIZE,
                    tab_size=tab_size,
                ):
                    self.assertEqual(
                        normal.get_expanded_size(tab_size),
                        len(source.expandtabs(tab_size)),
                    )
                #
            #
        #


class LeadingWhitespace(TestCase):
    """LeadingWhitespace class"""

    def test_attributes_fuzzed(self) -> None:
        """initialization with various combinations of spaces and tabs"""
        for iteration in range(FUZZ_ITERATIONS):
            source = get_random_prefix()
            leading_whitespace = whitespace.LeadingWhitespace(source)
            display = ws2sym(source)
            with self.subTest(
                display,
                iteration=iteration,
                subject=KW_NUMBER_OF_TABS,
            ):
                self.assertEqual(leading_whitespace.explicit_tabs, source.count(TAB))
            #
            with self.subTest(
                display,
                iteration=iteration,
                subject=KW_TOTAL_NUMBER_OF_CHARS,
            ):
                self.assertEqual(len(leading_whitespace), len(source))
            #
            with self.subTest(
                display,
                iteration=iteration,
                subject=KW_LONGEST_RUN_OF_SPACES,
            ):
                self.assertEqual(
                    leading_whitespace.longest_run_of_spaces,
                    max(len(item) for item in source.split(TAB)),
                )
            #
            with self.subTest(
                display,
                iteration=iteration,
                subject=KW_NO_SBT,
            ):
                self.assertEqual(
                    leading_whitespace.no_spaces_before_tabs,
                    SP_TAB not in source,
                )
            #
        #

    def test_ge(self) -> None:
        """Test "greater or equal" capability"""
        first = "SSS"
        second = "TT"
        with self.subTest("by length (shortcut #1)", first=first, second=second):
            self.assertTrue(
                whitespace.LeadingWhitespace(sym2ws(first))
                >= whitespace.LeadingWhitespace(sym2ws(second))
            )
        #
        first = "TTS"
        second = "TSSS"
        with self.subTest(
            "by tab number if no sbt (shortcut #2)", first=first, second=second
        ):
            self.assertTrue(
                whitespace.LeadingWhitespace(sym2ws(first))
                >= whitespace.LeadingWhitespace(sym2ws(second))
            )
        #
        first = "STT"
        second = "STSS"
        with self.subTest(
            "normal evaluation if sbt (ge case)", first=first, second=second
        ):
            self.assertTrue(
                whitespace.LeadingWhitespace(sym2ws(first))
                >= whitespace.LeadingWhitespace(sym2ws(second))
            )
        #
        first = "ST"
        second = "STT"
        with self.subTest(
            "normal evaluation if sbt (lt resp not ge case)", first=first, second=second
        ):
            self.assertFalse(
                whitespace.LeadingWhitespace(sym2ws(first))
                >= whitespace.LeadingWhitespace(sym2ws(second))
            )
        #

    def test_ge_witnesses(self) -> None:
        """test the .greater_or_equal_witnesses() method"""
        with self.subTest(WG_GE_COMPATIBILITY):
            first = whitespace.LeadingWhitespace(
                sym2ws("SSSTT"), compatibility_mode=True
            )
            second = whitespace.LeadingWhitespace(
                sym2ws("SSSTSS"), compatibility_mode=True
            )
            self.assertEqual(
                str(first.greater_or_equal_witnesses(second)),
                f"{WG_GE_COMPATIBILITY}at tab sizes 2, 3, 4",
            )
        #
        with self.subTest(WG_GE_NEWSTYLE):
            first = whitespace.LeadingWhitespace(sym2ws("SSSTT"))
            second = whitespace.LeadingWhitespace(sym2ws("SSSTSS"))
            self.assertEqual(
                str(first.greater_or_equal_witnesses(second)),
                f"{WG_GE_NEWSTYLE}at tab sizes 2, 3, 4",
            )
        #

    def test_ne_witnesses(self) -> None:
        """test the .not_equal_witnesses() method"""
        with self.subTest(WG_NE_COMPATIBILITY):
            first = whitespace.LeadingWhitespace(
                sym2ws("SSSTT"), compatibility_mode=True
            )
            second = whitespace.LeadingWhitespace(
                sym2ws("SSSTSS"), compatibility_mode=True
            )
            self.assertEqual(
                str(first.not_equal_witnesses(second)),
                f"{WG_NE_COMPATIBILITY}at tab sizes 1, 3, 4",
            )
        #
        with self.subTest(WG_NE_NEWSTYLE):
            first = whitespace.LeadingWhitespace(sym2ws("SSSTT"))
            second = whitespace.LeadingWhitespace(sym2ws("SSSTSS"))
            self.assertEqual(
                str(first.not_equal_witnesses(second)),
                f"{WG_NE_NEWSTYLE}at tab sizes 1, 3, 4",
            )
        #



# -*- coding: utf-8 -*-

"""
Unit test the supertabnanny.parser module

"""

import io
import logging
import re
import sys
import tokenize

from unittest import TestCase, skipIf
from unittest.mock import patch

from supertabnanny import parser
from supertabnanny.whitespace import Witness, WitnessesGroup

from . import commons

ALL_EXAMPLES = commons.AllSources()

KW_ATTRIBUTE = "attribute"
KW_OUTPUT = "output"
KW_REDUCED_OUTPUT = "reduced output"

PARSER_ROOT_LOGGER = "supertabnanny.parser.ROOT_LOGGER"

# Python major and minor version as a 2-int tuple
PYTHON_MMVERSION = tuple(sys.version_info[:2])


def get_nannynag_re_pattern(
    lineno: int = 1,
    error_message: str = commons.EMPTY,
    tabsizes: str = commons.EMPTY,
    offending_line: str = commons.EMPTY,
):
    """Return a regular expression pattern for mathing a NannyNag exception"""
    return re.escape(
        f"""({lineno}, WitnessesGroup({error_message!r}) with tabsizes"""
        f""" [{tabsizes}], {offending_line!r})"""
    )


class NannyNag(TestCase):
    """NannyNag exception"""

    def test_attributes(self) -> None:
        """Initialization and attributes"""
        lineno, wg, full_line = 7, WitnessesGroup("dummy"), "\t   1 tab and 3 spaces"
        nag = parser.NannyNag(lineno, wg, full_line)
        with self.subTest(KW_ATTRIBUTE, name="lineno"):
            self.assertEqual(nag.lineno, lineno)
        #
        with self.subTest(KW_ATTRIBUTE, name="witnesses group"):
            self.assertEqual(repr(nag.witnesses), repr(wg))
        #
        with self.subTest(KW_ATTRIBUTE, name="full_line"):
            self.assertEqual(nag.full_line, full_line)
        #


class Diagnosis(TestCase):
    """Diagnosis base class"""

    def test_init_attrs(self) -> None:
        """Initialization and attributes"""
        filename = "test/example.py"
        diagnosis = parser.Diagnosis(filename)
        with self.subTest(KW_ATTRIBUTE, filename=filename):
            self.assertEqual(diagnosis.filename, filename)
        #
        with self.subTest(KW_ATTRIBUTE, lineno=-1):
            self.assertEqual(diagnosis.lineno, -1)
        #
        with self.subTest(KW_ATTRIBUTE, message=commons.EMPTY):
            self.assertEqual(diagnosis.message, commons.EMPTY)
        #
        with self.subTest(KW_ATTRIBUTE, detail=commons.EMPTY):
            self.assertEqual(diagnosis.detail, commons.EMPTY)
        #
        with self.subTest(KW_ATTRIBUTE, success=False):
            self.assertFalse(diagnosis.success)
        #
        with self.subTest(KW_ATTRIBUTE, tabsizes=[]):
            self.assertEqual(diagnosis.tabsizes, [])
        #

    def test_qf_property(self) -> None:
        """.quoted_filename property"""
        filename = "test/example with whitespace.py"
        expected_qf = "'test/example with whitespace.py'"
        diagnosis = parser.Diagnosis(filename)
        with self.subTest(KW_ATTRIBUTE, expected_qf=expected_qf):
            self.assertEqual(diagnosis.quoted_filename, expected_qf)
        #

    def test_compatible_output(self) -> None:
        """.compatible_output() function"""
        for filename, expected_result in (
            ("test/example.py", "'test/example.py': : "),
            (
                "test/example with whitespace.py",
                "'test/example with whitespace.py': : ",
            ),
        ):
            with self.subTest(KW_OUTPUT, expected_result=expected_result):
                diagnosis = parser.Diagnosis(filename)
                self.assertEqual(diagnosis.compatible_output(), expected_result)
            #
        #

    def test_json_output(self) -> None:
        """.json_output() function"""
        for filename, expected_result in (
            (
                "test/example.py",
                '{"filename": "test/example.py", "lineno": -1, "message": "",'
                ' "detail": "", "success": false, "tabsizes": []}',
            ),
            (
                "test/example with whitespace.py",
                '{"filename": "test/example with whitespace.py", "lineno": -1,'
                ' "message": "", "detail": "", "success": false, "tabsizes": []}',
            ),
        ):
            with self.subTest(KW_OUTPUT, expected_result=expected_result):
                diagnosis = parser.Diagnosis(filename)
                self.assertEqual(diagnosis.json_output(), expected_result)
            #
        #

    def test_normal_output(self) -> None:
        """.normal_output() function"""
        for filename, expected_result in (
            ("test/example.py", "test/example.py: "),
            (
                "test/example with whitespace.py",
                "'test/example with whitespace.py': ",
            ),
        ):
            with self.subTest(KW_OUTPUT, expected_result=expected_result):
                diagnosis = parser.Diagnosis(filename)
                self.assertEqual(diagnosis.normal_output(), expected_result)
            #
        #


class CleanDiagnosis(TestCase):
    """CleanDiagnosis class"""

    def test_init_attrs(self) -> None:
        """Initialization and attributes"""
        filename = "test/example.py"
        message = "good"
        diagnosis = parser.CleanDiagnosis(filename, message)
        with self.subTest(KW_ATTRIBUTE, filename=filename):
            self.assertEqual(diagnosis.filename, filename)
        #
        with self.subTest(KW_ATTRIBUTE, lineno=-1):
            self.assertEqual(diagnosis.lineno, -1)
        #
        with self.subTest(KW_ATTRIBUTE, message=message):
            self.assertEqual(diagnosis.message, message)
        #
        with self.subTest(KW_ATTRIBUTE, detail=commons.EMPTY):
            self.assertEqual(diagnosis.detail, commons.EMPTY)
        #
        with self.subTest(KW_ATTRIBUTE, success=True):
            self.assertTrue(diagnosis.success)
        #
        with self.subTest(KW_ATTRIBUTE, tabsizes=[]):
            self.assertEqual(diagnosis.tabsizes, [])
        #

    def test_compatible_output(self) -> None:
        """.compatible_output() function"""
        for filename, message, expected_result in (
            ("test/example.py", "good", "'test/example.py': good"),
            (
                "test/example with whitespace.py",
                "clean",
                "'test/example with whitespace.py': clean",
            ),
        ):
            diagnosis = parser.CleanDiagnosis(filename, message)
            with patch(PARSER_ROOT_LOGGER, new=commons.MockLogger(level=logging.INFO)):
                with self.subTest(
                    KW_OUTPUT,
                    filename=filename,
                    message=message,
                    expected_result=expected_result,
                ):
                    self.assertEqual(diagnosis.compatible_output(), expected_result)
                #
            #
            with self.subTest(KW_REDUCED_OUTPUT, filename=filename, message=message):
                self.assertEqual(diagnosis.compatible_output(), commons.EMPTY)
            #
        #

    def test_normal_output(self) -> None:
        """.normal_output() function"""
        for filename, message, expected_result in (
            ("test/example.py", "good", "test/example.py: good"),
            (
                "test/example with whitespace.py",
                "clean",
                "'test/example with whitespace.py': clean",
            ),
        ):
            diagnosis = parser.CleanDiagnosis(filename, message)
            with patch(PARSER_ROOT_LOGGER, new=commons.MockLogger(level=logging.ERROR)):
                with self.subTest(
                    KW_REDUCED_OUTPUT, filename=filename, message=message
                ):
                    self.assertEqual(diagnosis.normal_output(), commons.EMPTY)
                #
            #
            with self.subTest(
                KW_OUTPUT,
                filename=filename,
                message=message,
                expected_result=expected_result,
            ):
                self.assertEqual(diagnosis.normal_output(), expected_result)
            #
        #


class ExceptionDiagnosis(TestCase):
    """ExceptionDiagnosis class"""

    def test_init_attrs(self) -> None:
        """Initialization and attributes"""
        filename = "test/example.py"
        syntax_error = SyntaxError("invalid", (filename, 7, 8, "pnirt", 7, 13))
        for exception, expected_lineno, expected_message, expected_detail in (
            (OSError("access forbidden"), -1, "I/O Error", "access forbidden"),
            (syntax_error, 7, "Syntax Error", str(syntax_error)),
            (ValueError("whatever"), -1, "ValueError", "whatever"),
        ):
            exc = str(exception)
            diagnosis = parser.ExceptionDiagnosis(filename, exc=exception)
            with self.subTest(KW_ATTRIBUTE, exc=exc, filename=filename):
                self.assertEqual(diagnosis.filename, filename)
            #
            with self.subTest(KW_ATTRIBUTE, exc=exc, expected_lineno=expected_lineno):
                self.assertEqual(diagnosis.lineno, expected_lineno)
            #
            with self.subTest(KW_ATTRIBUTE, exc=exc, expected_message=expected_message):
                self.assertEqual(diagnosis.message, expected_message)
            #
            with self.subTest(KW_ATTRIBUTE, exc=exc, detail=expected_detail):
                self.assertEqual(diagnosis.detail, expected_detail)
            #
            with self.subTest(KW_ATTRIBUTE, exc=exc, success=False):
                self.assertFalse(diagnosis.success)
            #
            with self.subTest(KW_ATTRIBUTE, exc=exc, tabsizes=[]):
                self.assertEqual(diagnosis.tabsizes, [])
            #
        #


class NagDiagnosis(TestCase):
    """NagDiagnosis class"""

    def test_init_attrs(self) -> None:
        """Initialization and attributes"""
        filename = "test/example.py"
        lineno = 67
        full_line = "           \t     \t  example content"
        witnesses_group = WitnessesGroup(prefix="multiple witnesses ")
        witnesses_group.add(Witness(4, 5, 6))
        witnesses_group.add(Witness(7, 8, 9))
        witnesses_group.add(Witness(20, 5, 5))
        diagnosis = parser.NagDiagnosis(
            filename, nag=parser.NannyNag(lineno, witnesses_group, full_line)
        )
        with self.subTest(KW_ATTRIBUTE, filename=filename):
            self.assertEqual(diagnosis.filename, filename)
        #
        with self.subTest(KW_ATTRIBUTE, lineno=lineno):
            self.assertEqual(diagnosis.lineno, lineno)
        #
        with self.subTest(KW_ATTRIBUTE, message="Line 67"):
            self.assertEqual(diagnosis.message, "Line 67")
        #
        with self.subTest(KW_ATTRIBUTE, detail=str(witnesses_group)):
            self.assertEqual(diagnosis.detail, str(witnesses_group))
        #
        with self.subTest(KW_ATTRIBUTE, success=False):
            self.assertFalse(diagnosis.success)
        #
        with self.subTest(KW_ATTRIBUTE, full_line=full_line):
            self.assertEqual(diagnosis.full_line, full_line)
        #
        with self.subTest(KW_ATTRIBUTE, tabsizes=[4, 7, 20]):
            self.assertEqual(diagnosis.tabsizes, [4, 7, 20])
        #

    def test_compatible_output(self) -> None:
        """.compatible_output() method"""
        filename = "test/example with spaces.py"
        lineno = 67
        full_line = "           \t     \t  example content"
        witnesses_group = WitnessesGroup(prefix="multiple witnesses ")
        witnesses_group.add(Witness(4, 5, 6))
        witnesses_group.add(Witness(7, 8, 9))
        witnesses_group.add(Witness(20, 5, 5))
        diagnosis = parser.NagDiagnosis(
            filename, nag=parser.NannyNag(lineno, witnesses_group, full_line)
        )
        with patch(PARSER_ROOT_LOGGER, new=commons.MockLogger(level=logging.ERROR)):
            with self.subTest(
                KW_REDUCED_OUTPUT,
                filename=filename,
            ):
                self.assertEqual(diagnosis.compatible_output(), f'"{filename}"')
            #
        #
        with patch(PARSER_ROOT_LOGGER, new=commons.MockLogger(level=logging.INFO)):
            with self.subTest(
                "full output",
                filename=filename,
            ):
                self.assertEqual(
                    diagnosis.compatible_output(),
                    f"{filename!r}: *** Line {lineno}: trouble in tab city! ***\n"
                    f"offending line: {full_line!r}\n"
                    "multiple witnesses at tab sizes 4, 7, 20",
                )
            #
        #
        with self.subTest("default output", filename=filename):
            self.assertEqual(
                diagnosis.compatible_output(), f'"{filename}" {lineno} {full_line!r}'
            )
        #


class IndentParserParseIO(TestCase):
    """IndentParser class, parse_io() method, tests modified from
    <https://github.com/python/cpython/blob/c501261/Lib/test/test_tabnanny.py>
    """

    def setUp(self) -> None:
        """Provide a parser"""
        self.i_parser = parser.IndentParser()
        # pylint: disable=invalid-name ; defined in unittest.TestCase
        self.maxDiff = None

    def _parse_literal(self, source_code: str) -> bool:
        r"""parse the _source\_code_ literal and return the result"""
        return self.i_parser.parse_io(io.StringIO(source_code))

    def _parse_source_codes(self, key: str) -> bool:
        r"""parse **SOURCE\_CODES\[_key_]** and return the result"""
        return self._parse_literal(ALL_EXAMPLES[key].source)

    def test_correct_source(self) -> None:
        """Python source code without any errors"""
        self.assertTrue(self._parse_source_codes(commons.ERROR_FREE))

    def test_wrong_indented(self) -> None:
        """Python source code eligible for raising an IndentationError"""
        self.assertRaisesRegex(
            IndentationError,
            "^unindent does not match any outer indentation level",
            self._parse_source_codes,
            commons.WRONG_INDENTED,
        )

    @skipIf(PYTHON_MMVERSION >= (3, 12), "up to 3.11 only")
    def test_tokenize_tokenerror_upto_311(self):
        """Python (up to 3.11) source code eligible for raising
        a tokenize.TokenError
        """
        expected_regex = re.escape("('EOF in multi-line statement', (9, 0))")
        self.assertRaisesRegex(
            tokenize.TokenError,
            f"^{expected_regex}$",
            self._parse_source_codes,
            commons.INCOMPLETE_EXPRESSION,
        )

    @skipIf(PYTHON_MMVERSION <= (3, 11), "from 3.12 only")
    def test_tokenize_tokenerror_from_312(self):
        """Python (from 3.12) source code eligible for raising
        a tokenize.TokenError
        """
        expected_regex = re.escape("('unexpected EOF in multi-line statement', (8, 0))")
        self.assertRaisesRegex(
            tokenize.TokenError,
            f"^{expected_regex}$",
            self._parse_source_codes,
            commons.INCOMPLETE_EXPRESSION,
        )

    @skipIf(PYTHON_MMVERSION >= (3, 12), "up to 3.11 only")
    def test_when_nannynag_error_upto_311(self):
        """Python (up to 3.11) source code eligible for raising
        a parser.NannyNag exception
        """
        expected_regex = get_nannynag_re_pattern(
            lineno=3,
            error_message="expanded size different ",
            tabsizes="1",
            offending_line='\tprint("world")\n',
        )
        self.assertRaisesRegex(
            parser.NannyNag,
            f"^{expected_regex}$",
            self._parse_source_codes,
            commons.NANNYNAG_ERRORED,
        )

    @skipIf(PYTHON_MMVERSION <= (3, 11), "from 3.12 only")
    def test_when_nannynag_error_from_312(self):
        """Python (from 3.12) source code eligible for raising
        a parser.NannyNag exception
        """
        expected_regex = get_nannynag_re_pattern(
            lineno=3,
            error_message=commons.TAB_ERROR_MESSAGE,
            tabsizes=commons.EMPTY,
            offending_line='\tprint("world")',
        )
        self.assertRaisesRegex(
            parser.NannyNag,
            f"^{expected_regex}$",
            self._parse_source_codes,
            commons.NANNYNAG_ERRORED,
        )

    @skipIf(PYTHON_MMVERSION >= (3, 12), "up to 3.11 only")
    def test_with_errored_codes_samples_upto311(self):
        """Python source code with whitespace related sampled problems
        with Python 3.11 or lower:

        *   "tab_space_errored_1": matched by the block under

                elif current_token.type == tokenize.INDENT

            in the .feed() method.
        *   "tab space_errored_2": matched block under

                elif self.check_is_required
                and current_token.type not in self.ignored_item

            in the .feed() method`
        """
        for source_key, error_message, tabsizes, offending_line in (
            (
                commons.TAB_SPACE_ERRORED_1,
                "expanded size greater or equal ",
                "1, 2",
                '\t\tprint("If called")\n',
            ),
            (
                commons.TAB_SPACE_ERRORED_2,
                "expanded size different ",
                "1, 2, 3, 4, 5, 6, 7, 9",
                '\t        print("If called")\n',
            ),
        ):
            with self.subTest(
                "NannyNag raised", source_key=source_key, offending_line=offending_line
            ):
                expected_regex = get_nannynag_re_pattern(
                    lineno=4,
                    error_message=error_message,
                    tabsizes=tabsizes,
                    offending_line=offending_line,
                )
                self.assertRaisesRegex(
                    parser.NannyNag,
                    f"^{expected_regex}$",
                    self._parse_source_codes,
                    source_key,
                )
            #
        #

    @skipIf(PYTHON_MMVERSION <= (3, 11), "from 3.12 only")
    def test_with_errored_codes_samples_from312(self):
        """Python source code with whitespace related sampled problems
        with Python 3.12 and up:

        *   "tab_space_errored_1": matched by the block under

                elif current_token.type == tokenize.INDENT

            in the .feed() method.
        *   "tab space_errored_2": matched block under

                elif self.check_is_required
                and current_token.type not in self.ignored_item

            in the .feed() method`
        """
        for source_key, offending_line in (
            (
                commons.TAB_SPACE_ERRORED_1,
                '\t\tprint("If called")',
            ),
            (
                commons.TAB_SPACE_ERRORED_2,
                '\t        print("If called")',
            ),
        ):
            with self.subTest(
                "NannyNag raised", source_key=source_key, offending_line=offending_line
            ):
                expected_regex = get_nannynag_re_pattern(
                    lineno=4,
                    error_message=commons.TAB_ERROR_MESSAGE,
                    tabsizes=commons.EMPTY,
                    offending_line=offending_line,
                )
                self.assertRaisesRegex(
                    parser.NannyNag,
                    f"^{expected_regex}$",
                    self._parse_source_codes,
                    source_key,
                )
            #
        #


class IndentParserParseFile(TestCase):
    """IndentParser class, parse_file() method, tests modified from
    <https://github.com/python/cpython/blob/c501261/Lib/test/test_tabnanny.py>
    """

    def setUp(self) -> None:
        """Provide a parser"""
        # self.i_parser = parser.IndentParser()
        self.compatibility_parser = parser.IndentParser(compatibility_mode=True)
        # pylint: disable=invalid-name ; defined in unittest.TestCase
        self.maxDiff = None

    def test_correct_source(self) -> None:
        """Python source code file without any errors"""
        diagnosis = self.compatibility_parser.parse_file(
            ALL_EXAMPLES[commons.ERROR_FREE].file_path
        )
        #
        with self.subTest("type"):
            self.assertIsInstance(diagnosis, parser.CleanDiagnosis)
        #
        with self.subTest("message"):
            self.assertEqual(diagnosis.message, "Clean bill of health.")
        #

    def test_wrong_indented_file(self) -> None:
        """Python source code file with indentation errors"""
        source_path = ALL_EXAMPLES[commons.WRONG_INDENTED].file_path
        diagnosis = self.compatibility_parser.parse_file(source_path)
        #
        with self.subTest("type"):
            self.assertIsInstance(diagnosis, parser.ExceptionDiagnosis)
        #
        with self.subTest("filename"):
            self.assertEqual(diagnosis.filename, str(source_path))
        #
        with self.subTest("message"):
            self.assertEqual(diagnosis.message, "Indentation Error")
        #
        error_source = "tokenize" if PYTHON_MMVERSION <= (3, 11) else "string"
        with self.subTest("detail"):
            self.assertEqual(
                diagnosis.detail,
                "unindent does not match any outer indentation level"
                f" (<{error_source}>, line 3)",
            )
        #

    @patch(PARSER_ROOT_LOGGER, new=commons.MockLogger(level=logging.INFO))
    def test_nannynag_error_verbose(self):
        """A python source code file eligible for raising parser.NannyNag.
        in verbose mode
        """
        example = ALL_EXAMPLES[commons.NANNYNAG_ERRORED]
        line_number = 3
        offending_line = example.get_line(line_number)
        diagnosis = self.compatibility_parser.parse_file(example.file_path)
        #
        reason = (
            "indent not equal e.g. at tab size 1"
            if PYTHON_MMVERSION <= (3, 11)
            else commons.TAB_ERROR_MESSAGE
        )
        self.assertEqual(
            diagnosis.compatible_output(),
            "\n".join(
                (
                    f"{str(example.file_path)!r}: *** Line {line_number}:"
                    " trouble in tab city! ***",
                    f"offending line: {offending_line!r}",
                    reason,
                )
            ),
        )

    def test_nannynag_error_normal(self):
        """A python source code file eligible for raising parser.NannyNag.
        in non-verbose mode
        """
        example = ALL_EXAMPLES[commons.NANNYNAG_ERRORED]
        line_number = 3
        offending_line = example.get_line(line_number)
        diagnosis = self.compatibility_parser.parse_file(example.file_path)
        self.assertEqual(
            diagnosis.compatible_output(),
            f"{example.file_path} {line_number} {offending_line!r}",
        )



# -*- coding: utf-8 -*-

"""
supertabnanny.parser - indent parser
"""

import dataclasses
import json
import logging
import pathlib
import shlex
import tokenize

from typing import TextIO

from .console import ROOT_LOGGER
from .whitespace import WitnessesGroup, LeadingWhitespace


EMPTY = ""
SPACE = " "
ZERO_WHITESPACE = LeadingWhitespace(EMPTY)


class NannyNag(Exception):
    """
    Raised by process_tokens() if detecting an ambiguous indent.
    Captured and handled in check().
    """

    def __init__(self, lineno: int, witnesses: WitnessesGroup, full_line: str) -> None:
        r"""Initialization arguments:

        *   _lineno_: line number where the error occurred
        *   _witnesses_: a whitespace.WitnessesGroup instance
        *   _full\_line_: the full line causing the error
        """
        self.lineno, self.witnesses, self.full_line = (
            lineno,
            witnesses,
            full_line,
        )


@dataclasses.dataclass
class Diagnosis:
    """Diagnosis base class"""

    filename: str
    _: dataclasses.KW_ONLY
    lineno: int = dataclasses.field(init=False, default=-1)
    message: str = dataclasses.field(init=False, default=EMPTY)
    detail: str = dataclasses.field(init=False, default=EMPTY)
    success: bool = dataclasses.field(init=False, default=False)
    tabsizes: list[int] = dataclasses.field(init=False, default_factory=list)

    @property
    def quoted_filename(self) -> str:
        """return shlex.quoted filename"""
        return shlex.quote(self.filename)

    def compatible_output(self) -> str:
        """Return output compatible to the original tabnanny"""
        return f"{self.filename!r}: {self.message}: {self.detail}"

    def json_output(self) -> str:
        """Return output as JSON"""
        return json.dumps(dataclasses.asdict(self))

    def normal_output(self) -> str:
        """Return normal human-readable output"""
        message = (
            f"{self.message}: {self.detail}"
            if self.detail and ROOT_LOGGER.level < logging.WARNING
            else self.message
        )
        return f"{self.quoted_filename}: {message}"


@dataclasses.dataclass
class CleanDiagnosis(Diagnosis):
    """Successful Diagnosis initialized with _filename_ and _message_"""

    message: str
    success: bool = dataclasses.field(init=False, default=True)

    def compatible_output(self) -> str:
        """Return output compatible to the original tabnanny if verbose"""
        if ROOT_LOGGER.level < logging.WARNING:
            return f"{self.filename!r}: {self.message}"
        #
        return EMPTY

    def normal_output(self) -> str:
        """Return normal human-readable output if not set to quiet mode"""
        if ROOT_LOGGER.level < logging.ERROR:
            return super().normal_output()
        #
        return EMPTY


@dataclasses.dataclass
class ErrorDiagnosis(Diagnosis):
    """Error Diagnosis base class"""


@dataclasses.dataclass
class ExceptionDiagnosis(ErrorDiagnosis):
    """Exception Diagnosis initialized with _filename_, and _exc_"""

    exc: dataclasses.InitVar[Exception] = dataclasses.InitVar(Exception)

    def __post_init__(self, exc: Exception):
        """Determine message by exception Type"""
        match exc:
            case OSError():
                self.message = "I/O Error"
            case tokenize.TokenError():
                self.message = "Token Error"
            case IndentationError():
                self.message = "Indentation Error"
            case SyntaxError():
                self.message = "Syntax Error"
            case _:
                self.message = f"{exc.__class__.__name__}"
            #
        #
        self.lineno = int(getattr(exc, "lineno", -1))
        self.detail = str(exc)


@dataclasses.dataclass
class NagDiagnosis(ErrorDiagnosis):
    """Nag Diagnosis class initialized with _filename_ and _nag_"""

    nag: dataclasses.InitVar[NannyNag] = dataclasses.InitVar(NannyNag)
    full_line: str = dataclasses.field(init=False)

    def __post_init__(self, nag: NannyNag):
        r"""Derive _message_, _detail_, _lineno_, and _full\_line_
        from  _nag_
        """
        self.message = f"Line {nag.lineno}"
        self.detail = str(nag.witnesses)
        self.lineno = nag.lineno
        self.full_line = nag.full_line
        self.tabsizes = nag.witnesses.tabsizes

    def compatible_output(self) -> str:
        """Return output compatible to the original tabnanny"""
        if ROOT_LOGGER.level < logging.WARNING:
            return (
                f"{self.filename!r}: *** {self.message}: trouble in tab city! ***\n"
                f"offending line: {self.full_line!r}\n{self.detail}"
            )
        #
        filename = self.filename
        if SPACE in filename:
            filename = f'"{self.filename}"'
        #
        if ROOT_LOGGER.level > logging.WARNING:
            return filename
        #
        return f"{filename} {self.lineno} {self.full_line!r}"


class IndentParser:
    """Parser for ..."""

    ignored_items = tokenize.COMMENT, tokenize.NL

    def __init__(self, compatibility_mode: bool = False) -> None:
        """Initialization argument:
        compatibility_mode: operate in full tabnanny compatibility mode
        """
        self.__check_for_equal_indent: bool = False
        self.__active: bool = False
        self.__indents: list[LeadingWhitespace] = []
        self.__compatibility_mode = compatibility_mode

    @property
    def is_stopped(self) -> bool:
        """Signal whether the partser is in stopped state"""
        return not self.__active

    @property
    def check_is_required(self) -> bool:
        """Reveal the check flag"""
        return self.__check_for_equal_indent

    @property
    def previous_indent(self) -> LeadingWhitespace:
        """Return the last indent from the indents stack"""
        return self.__indents[-1]

    def reset(self) -> None:
        """Reset internal state and signal readiness"""
        self.avoid_check()
        self.__indents.clear()
        self.push_indent(ZERO_WHITESPACE)
        self.__active = True

    def stop(self) -> None:
        """Stop the parser: clear indents cache and set not-ready"""
        self.__indents.clear()
        self.__active = False

    def avoid_check(self) -> None:
        """Unset the check flag"""
        self.__check_for_equal_indent = False

    def require_check(self) -> None:
        """Set the check flag"""
        self.__check_for_equal_indent = True

    def push_indent(self, leading_whitespace: LeadingWhitespace) -> None:
        """Push leading_whitespace to the indents stack"""
        logging.debug("pushing %r", leading_whitespace.raw)
        logging.debug("frequencies: %r", leading_whitespace.normalization.frequencies)
        self.__indents.append(leading_whitespace)

    def pop_indent(self) -> None:
        """remove the last entry from the indents stack"""
        last_raw_entry = self.__indents.pop().raw
        logging.debug("popped %r", last_raw_entry)

    def feed(self, current_token: tokenize.TokenInfo) -> None:
        """Feed a token to the parser"""
        if self.is_stopped:
            raise ValueError("Parser is stopped and cannot be fed any tokens")
        #
        previous_lw = self.previous_indent
        if current_token.type == tokenize.NEWLINE:
            self.require_check()
        elif current_token.type == tokenize.INDENT:
            self.avoid_check()
            current_lw = LeadingWhitespace(
                current_token.string, compatibility_mode=self.__compatibility_mode
            )
            if previous_lw >= current_lw:
                raise NannyNag(
                    current_token.start[0],
                    previous_lw.greater_or_equal_witnesses(current_lw),
                    current_token.line,
                )
            #
            self.push_indent(current_lw)
        elif current_token.type == tokenize.DEDENT:
            self.require_check()
            self.pop_indent()
        elif self.check_is_required and current_token.type not in self.ignored_items:
            self.avoid_check()
            current_lw = LeadingWhitespace(
                current_token.line, compatibility_mode=self.__compatibility_mode
            )
            if previous_lw != current_lw:
                raise NannyNag(
                    current_token.start[0],
                    previous_lw.not_equal_witnesses(current_lw),
                    current_token.line,
                )
            #
        #

    def parse_io(self, text_io: TextIO) -> bool:
        """Parse TextIO"""
        self.reset()
        try:
            for token_info in tokenize.generate_tokens(text_io.readline):
                self.feed(token_info)
            #
        except TabError as error:
            raise NannyNag(
                error.lineno or -1, WitnessesGroup(error.msg), error.text or "unknown"
            ) from error
        except (tokenize.TokenError, IndentationError, SyntaxError, NannyNag) as error:
            raise error from error
        finally:
            self.stop()
        #
        return True

    def parse_file(
        self,
        file_spec: str | pathlib.Path,
    ) -> Diagnosis:
        """Parse a single file and return a Diagnosis subclass"""
        if isinstance(file_spec, pathlib.Path):
            file_path = file_spec
        else:
            file_path = pathlib.Path(file_spec)
        #
        file_name = str(file_path)
        try:
            with tokenize.open(file_spec) as source_file:
                self.parse_io(source_file)
            #
        except (OSError, tokenize.TokenError, IndentationError, SyntaxError) as exc:
            return ExceptionDiagnosis(file_name, exc=exc)
        except NannyNag as nag:
            return NagDiagnosis(file_name, nag=nag)
        #
        return CleanDiagnosis(file_name, message="Clean bill of health.")
